# Author: Carsten Sachse 07-Dec-2010
# Copyright: EMBL (2010 - 2018)
# License: see license.txt for details
"""
Program to split large scan into individual micrographs. The location of 
individual micrographs is determined by a dummy micrograph reference.
"""
from EMAN2 import EMData, Util, EMNumPy
from collections import namedtuple
from fundamentals import ccfn, image_decimate, mirror
from morphology import threshold_outside
from spring.csinfrastr.csfeatures import Features, FeaturesSupport
from spring.csinfrastr.cslogger import Logger
from spring.csinfrastr.csproductivity import DiagnosticPlot, Temporary, OpenMpi
from spring.csinfrastr.csreadinput import OptHandler
from utilities import peak_search, model_blank
import matplotlib.cm as cm
import numpy as np
import os
import sys


class Micrograph(object):
    def readmic(self, inputimg):
        """
        * Function to load micrograph and define sizes

        #. Usage: img, nx, ny = readmic(inputimg)
        #. Input: inputimg
        #. Output: image, x and y dimension

        """
        log = Logger()
        log.fcttolog()

        img = EMData()
        img.read_image(inputimg)
        log.ilog('Input micrograph read: %s' %inputimg)

        nx = img.get_xsize()
        ny = img.get_ysize()
        
        return img, nx, ny


    def get_statistics_from_image(self, image, mask=None):
        mean, sigma, imin, imax = Util.infomask(image, mask, True)
        img_stat = namedtuple('image_statistics', 'avg sigma min max')
        statistics = img_stat(mean, sigma, imin, imax)
        
        return statistics
    

    def adjust_gray_values_for_print_and_optimal_display(self, mic):
        stat = self.get_statistics_from_image(mic)
        min_cutoff, max_cutoff = stat.avg - 3.5 * stat.sigma, stat.avg + 3.5 * stat.sigma
        pix_range = stat.max - stat.min
        if stat.avg < stat.min + 0.1 * pix_range:
            min_cutoff = stat.min
        elif stat.avg > stat.min + 0.9 * pix_range:
            max_cutoff = stat.max
            
        adjusted_mic = threshold_outside(mic, min_cutoff, max_cutoff)
        
        return adjusted_mic
    
        
class ScanSplitPar:
    """
    Class to initiate default dictionary with input parameters including help and range values and 
    status dictionary
    """
    def __init__(self):
        # package/program identity
        self.package = 'emspring'
        self.progname = 'scansplit'
        self.proginfo = __doc__
        self.code_files = [self.progname, self.progname + '_mpi']

        self.scansplit_features = Features()
        self.feature_set = self.scansplit_features.setup(self)
        
        self.define_parameters_and_their_properties()
        self.define_program_states()


    def define_parameters_and_their_properties(self):

        self.feature_set = self.scansplit_features.set_inp_multiple_micrographs(self.feature_set)
        self.feature_set = self.set_mic_out_pattern(self.feature_set)
    
        self.feature_set = self.set_rows_and_cols_count(self.feature_set)
        self.feature_set = self.set_mic_crop_count(self.feature_set)
        self.feature_set = self.set_mic_width_and_height(self.feature_set)
        self.feature_set = self.set_label_width_and_height(self.feature_set)
        self.feature_set = self.scansplit_features.set_scanner_step_size(self.feature_set)
        self.feature_set = self.set_cc_threshold(self.feature_set)
        self.feature_set = self.set_contact_print_option(self.feature_set)
        self.feature_set = self.set_final_print_option(self.feature_set)
        self.feature_set = self.set_normscan_option(self.feature_set)
    
        self.feature_set = self.scansplit_features.set_binning_option(self.feature_set)
        self.feature_set = self.scansplit_features.set_binning_factor(self.feature_set)
        
        self.feature_set = self.scansplit_features.set_mpi(self.feature_set)
        self.feature_set = self.scansplit_features.set_ncpus_scan(self.feature_set)
        self.feature_set = self.scansplit_features.set_temppath(self.feature_set)
    
    def define_program_states(self):
        # status dictionary
        self.feature_set.program_states['readmic']='Read input micrograph'
        self.feature_set.program_states['roughsplit']='Rough splitting of micrograph according to number of rows ' +\
        'and columns'
        self.feature_set.program_states['make_refmic']='Make binary reference micrograph including label'
        self.feature_set.program_states['findlabel']='Find label on micrograph with respect to reference micrograph'
        self.feature_set.program_states['finesplit']='Fine splitting of micrograph according to matched position'
        self.feature_set.program_states['normscan']='Normalize micrograph columns using Niko Grigorieff\'s Normscan'
        self.feature_set.program_states['bin_micrographs']='Bin micrographs by specified factor'
        self.feature_set.program_states['contactprint']='Produce contact print of scan'
        self.feature_set.program_states['finalprint']='Produce final prints of split micrograph'

    def set_mic_out_pattern(self, feature_set):
        inp2 = 'Output micrograph pattern'
        feature_set.parameters[inp2] = 'cs_scan034???.mrc'
        
        feature_set.properties[inp2] = feature_set.file_properties(1, ['jpg', 'tif', 'mrc', 'spi', 'hdf', 'img', 'hed'],
        'saveFile')
        
        feature_set.hints[inp2] = 'If single input micrograph: name of pattern. In case of multiple micrographs ' + \
        'suffix to be attached to corresponding micrograph. Use wildcards cs_scan034\?\?\?.mrc OR cs_scan034*.mrc, ' + \
        FeaturesSupport().add_accepted_file_formats_to_hint(feature_set, inp2)
        feature_set.level[inp2]='beginner'
        
        return feature_set


    def set_rows_and_cols_count(self, feature_set):
        inp4 = 'Number of columns x rows'
        feature_set.parameters[inp4] = tuple((3, 2))
        feature_set.hints[inp4] = 'Integer value for number of columns x rows of micrographs located on scan.'
        feature_set.properties[inp4] = feature_set.Range(1, 10, 1)
        feature_set.level[inp4]='beginner'
        
        return feature_set


    def set_mic_crop_count(self, feature_set):
        inp5 = 'Number of micrographs to be cropped'
        feature_set.parameters[inp5] = int(6)
        feature_set.hints[inp5] = 'Integer value number of micrographs located on scan (number of rows x number of ' +\
        'columns).'
        feature_set.properties[inp5] = feature_set.Range(1, 100, 1)
        feature_set.level[inp5]='beginner'
        
        return feature_set


    def set_mic_width_and_height(self, feature_set):
        inp6 = 'Micrograph width x height in cm'
        feature_set.parameters[inp6] = tuple((8.0, 9.0))
        feature_set.hints[inp6] = 'Width x height dimension of micrograph in cm.'
        feature_set.properties[inp6] = feature_set.Range(0, 50, 0.1)
        feature_set.level[inp6]='intermediate'
        
        return feature_set


    def set_label_width_and_height(self, feature_set):
        inp7 = 'Label width x height in cm'
        feature_set.parameters[inp7] = tuple((2.7, 1.4))
        feature_set.hints[inp7] = 'Width x height dimension of black label in cm located at bottom of micrograph.'
        feature_set.properties[inp7] = feature_set.Range(0, 10, 0.1)
        feature_set.level[inp7]='expert'
        
        return feature_set


    def set_cc_threshold(self, feature_set):
        inp9 = 'Cross-correlation rejection criterion'
        feature_set.parameters[inp9] = float(1e+4)
        feature_set.hints[inp9] = 'Use the default number to exclude empty positions with no micrograph on scan - ' + \
        'otherwise use with caution.'
        feature_set.properties[inp9] = feature_set.Range(0, 1e+8, 1e+3)
        feature_set.level[inp9]='expert'
        
        return feature_set


    def set_contact_print_option(self, feature_set):
        inp11 = 'Contact print option'
        feature_set.parameters[inp11] = bool(False)
        feature_set.hints[inp11] = 'Option to produce a contact print of the entire scan.'
        feature_set.level[inp11]='intermediate'
        
        return feature_set


    def set_final_print_option(self, feature_set):
        inp12 = 'Final print option'
        feature_set.parameters[inp12] = bool(False)
        feature_set.hints[inp12] = 'Option to produce a final high-contrast print of the cropped micrograph.'
        feature_set.level[inp12]='intermediate'
        
        
        return feature_set


    def set_normscan_option(self, feature_set):
        inp13 = 'Normscan option'
        feature_set.parameters[inp13] = bool(False)
        feature_set.hints[inp13] = 'Option to normalize micrograph columns to eliminate steep grey-value differences '+\
        'in micrograph columns that can be present in Zeiss scans.'
        feature_set.level[inp13]='intermediate'
        
        return feature_set
    

class ScanSplitRough(object):
    """
    * Class that holds all functions required for splitting micrographs 

    * __init__ Function to read in the entered parameter dictionary, load micrograph and initialize \
                    unique temporary directory

    #. Usage: ScanSplit(pardict)
    #. Input: pardict = OrderedDict of program parameters
    """
    def __init__(self, parset = None):
        self.log = Logger()
        if parset is not None: 
            self.feature_set = parset
            p = self.feature_set.parameters
            self.infile=p['Micrographs']
            self.micrograph_files = Features().convert_list_of_files_from_entry_string(self.infile)
            self.outfile=p['Output micrograph pattern']
            self.column_count=p['Number of columns x rows'][0]
            self.row_count=p['Number of columns x rows'][1]
            self.mic_count=p['Number of micrographs to be cropped']
            self.mwidth=p['Micrograph width x height in cm'][0]
            self.mheight=p['Micrograph width x height in cm'][1]
            self.lwidth=p['Label width x height in cm'][0]
            self.lheight=p['Label width x height in cm'][1]
            self.scanstep=p['Scanner step size in micrometer']
            self.cc_threshold=p['Cross-correlation rejection criterion']

            self.normscanoption=p['Normscan option']
            self.binoption=p['Binning option']
            self.binfactor=p['Binning factor']
            if self.binfactor == 1 and self.binoption is True:
                self.binoption = False

            self.cprintoption=p['Contact print option']
            self.fprintoption=p['Final print option']

            self.temppath=p['Temporary directory']
            self.mpi_option = p['MPI option']
            self.cpu_count = p['Number of CPUs']


    def roughsplit(self, scanimg, row_count, column_count, no_mics):
        """
        * Function to roughly split scan into micrographs based on given number \
        of rows and columns

        #. Usage: roughsplit(row_count, column_count, mic_count)
        #. Input: number of rows, number of columns, number of micrographs, 
        #. Output: dictionary of roughly windowed micrographs (EMData objects)

        >>> from spring.micprgs.scansplit import ScanSplit
        >>> import numpy as np
        >>> from EMAN2 import EMNumPy
        >>> scanimg = EMNumPy.numpy2em(np.reshape(np.arange(0, 100), (20, 5)))
        >>> mic_one, decfactor = ScanSplit().roughsplit(scanimg, 2, 1, 2)
        >>> np.rint(EMNumPy.em2numpy(mic_one[0]))
        array([[  0.,   1.,   2.,   3.,   4.],
               [  5.,   6.,   7.,   8.,   9.],
               [ 10.,  11.,  12.,  13.,  14.],
               [ 15.,  16.,  17.,  18.,  19.],
               [ 20.,  21.,  22.,  23.,  24.],
               [ 25.,  26.,  27.,  28.,  29.],
               [ 30.,  31.,  32.,  33.,  34.],
               [ 35.,  36.,  37.,  38.,  39.],
               [ 40.,  41.,  42.,  43.,  44.],
               [ 45.,  46.,  47.,  48.,  49.]], dtype=float32)

        >>> scanimg = EMNumPy.numpy2em(np.reshape(np.arange(0, 100), (10, 10)))
        >>> mic_one, decfactor = ScanSplit().roughsplit(scanimg, 1, 2, 2)
        >>> np.rint(EMNumPy.em2numpy(mic_one[0]))
        array([[  0.,   1.,   2.,   3.,   4.],
               [ 10.,  11.,  12.,  13.,  14.],
               [ 20.,  21.,  22.,  23.,  24.],
               [ 30.,  31.,  32.,  33.,  34.],
               [ 40.,  41.,  42.,  43.,  44.],
               [ 50.,  51.,  52.,  53.,  54.],
               [ 60.,  61.,  62.,  63.,  64.],
               [ 70.,  71.,  72.,  73.,  74.],
               [ 80.,  81.,  82.,  83.,  84.],
               [ 90.,  91.,  92.,  93.,  94.]], dtype=float32)
        """
        self.log.fcttolog()

        sizex = scanimg.get_xsize()/column_count
        sizey = scanimg.get_ysize()/row_count

        mic_number = 0 
        miclist = []
        decimation_factor_find = int(sizex/1200) + 1
        for each_row_number in range(row_count):
            for each_column_number in range(column_count):
                centerx = (-column_count + 1)*sizex/2 + each_column_number*sizex 
                centery = (-row_count + 1)*sizey/2 + each_row_number*sizey
                self.log.dlog('centerx: %g, centery: %g'%(centerx, centery))
                splitmic = Util.window(scanimg, sizex, sizey, 1, centerx, centery, 0)
                # bin images to reduce computing time significantly
                if decimation_factor_find > 1:
                    splitmic = image_decimate(splitmic, decimation_factor_find, fit_to_fft=False)
                    self.log.dlog('micgraph was decimated by {0}'.format(decimation_factor_find))
                miclist.append(splitmic)
                self.log.ilog('Micrograph %d roughly windowed' %mic_number)
                mic_number = mic_number + 1

        return miclist, decimation_factor_find
    
    
class ScanSplitFind(ScanSplitRough):

    def make_refmic(self, mwidth = None, mheight = None, lwidth = None, lheight = None, scanstep = None):
        """
        * Function to generate a reference micrograph including black label at \
        bottom with the given dimensions

        #. Usage: make_refmic(mwidth, mheight, lwidth, lheight, scanstep)
        #. Input: micrograph widht, micrograph height, label width, label \
        height, scanner stepsize
        #. Output: reference micrograph

        >>> from spring.micprgs.scansplit import ScanSplit
        >>> refmic = ScanSplit().make_refmic(9, 10, 3, 2, 10000)
        >>> import numpy as np
        >>> from EMAN2 import EMNumPy
        >>> np.rint(EMNumPy.em2numpy(refmic))
        array([[ 1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.],
               [ 1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.],
               [ 1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.],
               [ 1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.],
               [ 1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.],
               [ 1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.],
               [ 1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.],
               [ 1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.],
               [ 1.,  1.,  1.,  0.,  0.,  0.,  1.,  1.,  1.],
               [ 1.,  1.,  1.,  0.,  0.,  0.,  1.,  1.,  1.]], dtype=float32)
  
        """
        self.log.fcttolog()
          
        if mwidth is None: mwidth = self.mwidth
        if mheight is None: mheight = self.mheight
        if lwidth is None: lwidth = self.lwidth
        if lheight is None: lheight = self.lheight
        if scanstep is None: scanstep = self.scanstep

        labelx = int(lwidth*1e+4/(scanstep))
        labely = int(lheight*1e+4/(scanstep))
        labelimg = model_blank(labelx, labely)

        mic_width = int(mwidth*1e+4/(scanstep))
        mic_height = int(mheight*1e+4/(scanstep))

        try:
            refmic = Util.pad(labelimg, mic_width, mic_height, 1, 
                                   0, int(round(mic_height/2 - labely/2)), 0, '1')
        except:
            refmic = Util.pad(labelimg, mic_width, mic_height, 1, 
                                   0, int(round(mic_height/2 - labely/2 - 1)), 0, '1')

        return refmic
    

    def adjust_refmic(self, refmic, miclist):
        """
        * Function to adjust size of reference microgrph to the size of rough split
        """

        mic = miclist[0]
        nx = int(mic.get_xsize())
        ny = int(mic.get_ysize())
        self.log.ilog('Pad label to %d, %d' %(nx, ny))

        # pad reference micrograph
        labelimg = refmic
        self.log.dlog('labelimg of %g and %g to be padded to: %g, %g' %
                      (labelimg.get_xsize(), labelimg.get_ysize(), nx, ny))
        try:
            labelpd = Util.pad(labelimg, nx, labelimg.get_ysize(), 1, 0, 0, 0, '0')
        except:
            labelpd = Util.window(labelimg, nx, labelimg.get_ysize(), 1, 0, 0, 0)
    
        try:
            labelpd = Util.pad(labelpd, nx, ny, 1, 0, 0, 0, '0')
        except:
            labelpd = Util.window(labelpd, nx, ny, 1, 0, 0, 0)
            
        return labelpd
    

    def determine_micrograph_offset_against_reference_micrograph(self, refmic_with_label_top, refmic_with_label_bottom,
    micrograph):
        mirror_ccpeaks = []
        mirror_xoffset = []
        mirror_yoffset = []
        x_dimension = int(micrograph.get_xsize())
        for reference_micrograph_with_label in [refmic_with_label_top, refmic_with_label_bottom]:
            ccimg = ccfn(micrograph, reference_micrograph_with_label, center=True) 
            ccimg = Util.window(ccimg, int(0.5 * x_dimension), int(0.5 * x_dimension), 1, 0, 0, 0) #
            peakcc = peak_search(ccimg, 1, +1)
            ccvv, nnn, nnn, nnn, ccxx, ccyy = peakcc[0]
            mirror_ccpeaks.append(ccvv)
            mirror_xoffset.append(ccxx)
            mirror_yoffset.append(ccyy)
        
        if mirror_ccpeaks[0] == max(mirror_ccpeaks):
            ccvv = mirror_ccpeaks[0]
            ccxx = mirror_xoffset[0]
            ccyy = mirror_yoffset[0]
            
        return ccvv, ccxx, ccyy
    

    def findlabel(self, miclist, decimation_factor_find, scanstep):
        """
        * Function to locate micrograph position using cross-correlation map with reference micrograph

        #. Usage: findlabel(refmic, miclist)
        #. Input: stack of roughly split micrographs, reference micrograph, number of micrographs,\
        dictionary that holds micrograph objects, number of micrographs
        #. Output: Position of X-offset and Y-offset and cross-correlation peak value of match
        """
        self.log.fcttolog()

        refmic = self.make_refmic(scanstep=scanstep*decimation_factor_find)

        refmic_with_label_top = self.adjust_refmic(refmic, miclist)
        refmic_with_label_bottom = mirror(refmic_with_label_top, axis='y')
        
        ccpeaks = []
        xoffset = []
        yoffset = []
        for each_mic_number, each_micrograph in enumerate(miclist):
            ccvv, ccxx, ccyy = self.determine_micrograph_offset_against_reference_micrograph(refmic_with_label_top, 
                                                                        refmic_with_label_bottom, each_micrograph)
                
            ccpeaks.append(ccvv)
            xoffset.append(ccxx*decimation_factor_find)
            yoffset.append(ccyy*decimation_factor_find)
            
            self.log.ilog('Micrograph center was found at %d, %d pixels with respect to reference \
                            micrograph (%g cc-value)' %
                            (xoffset[each_mic_number], yoffset[each_mic_number], ccpeaks[each_mic_number]))

        return xoffset, yoffset, ccpeaks
    

class ScanSplitOptions(ScanSplitFind):

    def assign_edge_value_to_edge_columns(self, column_number, column_count):
        """
        >>> ScanSplit().assign_edge_value_to_edge_columns(2, 100)
        6
        >>> ScanSplit().assign_edge_value_to_edge_columns(66, 100)
        66
        >>> ScanSplit().assign_edge_value_to_edge_columns(95, 100)
        94
        """
        if column_number < 0.06 * column_count:
            reference_column_number = int(0.06 * column_count)
        elif column_number > 0.94 * column_count:
            reference_column_number = int(0.94 * column_count)
        else:
            reference_column_number = column_number
        
        return reference_column_number
            

    def normalize_columns(self, micrograph):
        """
        determine number of columns
        pull out each column
        normalize
        """
        column_count = micrograph.get_xsize()
        row_count = int(0.5*micrograph.get_ysize())

        micrograph_center = Util.window(micrograph, column_count, row_count, 1, 0, 0, 0)

        for column_number in range(column_count):
            reference_column_number = self.assign_edge_value_to_edge_columns(column_number, column_count)
            reference_column = micrograph_center.get_col(reference_column_number)
            mean, sigma, imin, imax = Micrograph().get_statistics_from_image(reference_column)
            normalized_column = (micrograph.get_col(column_number) - mean) / sigma

            micrograph.set_col(normalized_column, column_number)
        
        return micrograph
            

    def normscan(self, miclist):
        """
        * Function to normalize micrograph columns from Zeiss scans

        #. Usage: normscan(normscanpath, miclist)
        #. Input: list of micrographs
        #. Output: Series of normalized micrographs, list of normalized micrograph filenames

        """
        self.log.fcttolog()

        normmiclist = []
        micrograph = EMData()
        for each_micfile in miclist:
            self.log.ilog('Normalize micrograph {0} columns.'.format(each_micfile))
            micrograph.read_image(each_micfile)
            self.micrograph = self.normalize_columns(micrograph)

            self.micrograph.write_image(each_micfile)
            normmiclist.append(each_micfile)
            
        return normmiclist
    

    def contactprint(self, scanimg, infile):
        """
        * Function to generate a contact print of the negative scan 

        #. Usage: contactprint(scanimg, infile) 
        #. Input: micrograph EMData object, input filename 
        #. Output: compressed contact print ( file)
        """
        
        outfile = '{0}_cprint{1}pdf'.format(os.path.splitext(os.path.basename(infile))[0], os.extsep)
        fig = self.prepare_print(scanimg, infile, outfile)
        fig.savefig(outfile, dpi=300, orientation='portrait')


    def prepare_print(self, img, infile, outfile, fig_number=None):
        """
        * Function to write print

        #. Usage: fig = prepare_print(img, infile, outfile, fig_number)
        #. Input: image to be printed, input filename, outputfilename, figure number
        #. Output: figure to be saved
        """

        split_micrograph_plot = DiagnosticPlot()
        fig = split_micrograph_plot.add_header_and_footer(self.feature_set, infile, outfile)

        ax1 = fig.add_subplot(111)
        if img.get_xsize() > 5000:
            decimation_factor = int(img.get_xsize()/5000) + 1
            img = Util.decimate(img, decimation_factor, decimation_factor, 1)
            self.log.dlog('Image {0} was binned by {1}'.format(outfile, decimation_factor))
            
        img_array = np.copy(EMNumPy.em2numpy(img))
        self.log.dlog('Image {0} was converted to Numpy array'.format(infile))

        height, width = img_array.shape
        if height > width:
            img_array = np.swapaxes(img_array, 0, 1)
        ax1.imshow(img_array, cmap=cm.gray, interpolation='nearest')

        return fig


    def finalprint(self, miclist):
        """
        * Function to generate final prints of the cropped micrographs 

        #. Usage: finalprint(miclist, scanstep, tempdir)
        #. Input: list of micrographs, scanner stepsize, temporary directory
        #. Output: series of final prints (compressed JPG file)
        """

        splitmic = EMData()
        for filenumber, infile in enumerate(miclist):
            outfile = '{0}_fprint{1}pdf'.format(infile.split(os.extsep)[0], os.extsep)
            splitmic.read_image(infile)
            splitmic = Micrograph().adjust_gray_values_for_print_and_optimal_display(splitmic)

            fig = self.prepare_print(splitmic, infile, outfile, filenumber)
            self.log.ilog('Micrograph {0} saved.'.format(outfile))
            fig.savefig(outfile, dpi=300, orientation='portrait')
            
        
    def bin_micrographs(self, miclist, binfactor):
        """
        * Function to bin series of micrographs by binfactor

        #. Usage: bin_micrographs(miclist, binfactor)
        #. Input: list of micrograph names, binfactor (1x = nobinning)
        #. Output: micrograph names will be appended by -binfactorxbinned.ext
        """
        self.log.fcttolog()

        mics_binned = []
        if binfactor < 2:
            sys.stderr.write('Bin factor of %g too low, binning is not required' %binfactor)
        else:
            micg = EMData()
            for mic in miclist: 
                micbin = mic.split(os.extsep)
                micg.read_image(mic)
                micg = image_decimate(micg, binfactor, fit_to_fft=False)
                micbinned = micbin[0] + '-' + '%d' %binfactor + 'xbin' + os.extsep + micbin[-1]
                micg.write_image(micbinned)
                mics_binned.append(micbinned)
                self.log.ilog('Micrograph %s was binned to %s.' %(mic, micbinned))
        
        return mics_binned
    

class ScanSplit(ScanSplitOptions):
    def split_output_string(self, outfile):
        """
        >>> ScanSplit().split_output_string('test_???.mrc')
        ['test_', '.mrc']
        >>> ScanSplit().split_output_string('test_??.mrc')
        ['test_', '.mrc']
        >>> ScanSplit().split_output_string('test_?.mrc')
        ['test_', '.mrc']
        >>> ScanSplit().split_output_string('test_*.mrc')
        ['test_', '.mrc']
        >>> ScanSplit().split_output_string('test_%02d.mrc')
        ['test_', '.mrc']
        """
#        if outfile is None: outfile = self.outfile

        if '???' in outfile: 
            output_split = outfile.split('???')
        elif '??' in outfile: 
            output_split = outfile.split('??')
        elif '?' in outfile: 
            output_split = outfile.split('?')
        elif '*' in outfile: 
            output_split = outfile.split('*')
        elif '%02d' in outfile: 
            output_split = outfile.split('%02d')
        else: 
            output_split = outfile.split(os.extsep)
            output_split[-2] = output_split[-2] + '_'
            output_split[-1] = os.extsep + output_split[-1]

        return output_split
    

    def compute_micrograph_center_in_scan(self, col_row_number, col_row_count, size, offset, mic_number):
        column_number, row_number = col_row_number
        column_count, row_count = col_row_count
        xoffset, yoffset = offset
        sizex, sizey = size
        centerx = (-column_count + 1)*sizex/2 + \
                        column_number*sizex + xoffset[mic_number]
        centery = (-row_count + 1)*sizey/2 + \
                        row_number*sizey + yoffset[mic_number]
        self.log.dlog('Micrograph\'s off-center position: {0} and {1}'.format(centerx, centery))
        
        return centerx, centery


    def check_and_adjust_windowing_paramters(self, scan_dimension, mic_dimension, center):
        """
        * Function to adjust windowing parameter in case of incompletely scanned micrographs
        
        >>> ScanSplit().check_and_adjust_windowing_paramters((1500, 1000), (500, 500), (300, 300))
        (500, 450, 300, 275)
        >>> ScanSplit().check_and_adjust_windowing_paramters((1000, 1000), (500, 500), (300, 300))
        (450, 450, 275, 275)
        >>> ScanSplit().check_and_adjust_windowing_paramters((1500, 1500), (500, 500), (300, 300))
        (500, 500, 300, 300)
        """
        nx, ny = scan_dimension
        mic_width, mic_height = mic_dimension
        centerx, centery = center
        
        newmic_width = mic_width
        newmic_height = mic_height
        if (abs(centerx) + mic_width/2 > nx/2): 
            overhang_x = abs(nx/2 - abs(centerx) - mic_width/2)
            newmic_width = mic_width - overhang_x
            if centerx < 0: 
                centerx = centerx + (mic_width - newmic_width)/2
            elif centerx > 0: 
                centerx = centerx - (mic_width - newmic_width)/2
                
            if newmic_width %2 != 0: 
                newmic_width = newmic_width - 1
            self.log.ilog('X-overhang is %g pixel while windowing. Micrograph size adjusted to %g x %g pixels' %
                          (overhang_x, newmic_width, newmic_height))

        if (abs(centery) + mic_height/2 > ny/2): 
            overhang_y = abs(ny/2 - abs(centery) - mic_height/2)
            newmic_height = mic_height - overhang_y
            if centery < 0: 
                centery = centery + (mic_height - newmic_height)/2
            elif centery > 0: 
                centery = centery - (mic_height - newmic_height)/2
            if newmic_height %2 != 0: 
                newmic_height = newmic_height - 1
            self.log.ilog('Y-overhang is %g pixel while windowing. Micrograph size adjusted to %g x %g pixels' %
                          (overhang_y, newmic_width, newmic_height))

        self.log.dlog('Windowing paramters: centerx: %g, centery: %g, newmic_width: %g, newmic_height: %g' %
                      (centerx, centery, newmic_width, newmic_height))
        
        return newmic_width, newmic_height, centerx, centery


    def finesplit(self, row_count, column_count, mwidth, mheight, scanstep, xoffset, yoffset, ccpeaks, cc_threshold,
    outfile):
        """
        * Function to finely split scan into micrographs according to found micrograph location

        #. Usage: finesplit(row_count, column_count, mwidth, mheight, scanstep, xoffset,\
        yoffset, ccpeaks, cc_threshold, outfile)
        #. Input: number of rows, number of colums, number of micrographs,\
        micrograph width, micrograph height, scanner stepsize, X-offset center, Y-offset center,\
        list of cross-correlation peaks, cross-correlation threshold
        #. Output: series of cropped micrographs

        """
        xsize = self.nx/column_count
        ysize = self.ny/row_count

        mic_width = int(mwidth*10000/(scanstep))
        mic_height = int(mheight*10000/(scanstep))
        self.log.dlog('imgx: %g, imgy: %g sizex: %g, sizey: %g, mic_width: %g, mixy: %g' %
                      (self.nx, self.ny, xsize, ysize, mic_width, mic_height))

        mic_number = 0 
        miclist = []
        output_split = self.split_output_string(outfile)
        
        for each_row_number in range(row_count):
            for each_column_number in range(column_count):
                if ccpeaks[mic_number] > cc_threshold: 
                    centerx, centery = self.compute_micrograph_center_in_scan((each_column_number, each_row_number),
                    (column_count, row_count), (xsize, ysize), (xoffset, yoffset), mic_number)
                    
                    newmic_width, newmic_height, centerx, centery = self.check_and_adjust_windowing_paramters((self.nx,
                    self.ny), (mic_width, mic_height), (centerx, centery))
                    
                    if row_count == 1 and column_count == 1:
                        splitmic = self.img
                    else:
                        splitmic = Util.window(self.img, int(newmic_width), int(newmic_height), 1, int(centerx),
                        int(centery), 0)

                    fname = output_split[-2] + '%03d'%mic_number + output_split[-1]
                    self.log.ilog('Micrograph %s neatly windowed: %g x %g cm' %
                                  (fname, newmic_width*scanstep/10000, 
                                   newmic_height*scanstep/10000))
                    splitmic.write_image(fname)
                    miclist.append(fname)
                    mic_number = mic_number + 1
                else:
                    self.log.ilog('Micrograph {0} was NOT windowed because of low cross-correlation. '.format(mic_number) + \
                    'If this is not correct, lower cross-correlation threshold')
        if mic_number is 0: 
            error_message = 'Exited because no micrographs found on scan (check cross-correlation '+ \
            'between dummy micrograph and set cc-threshold in parameters)'
            raise ValueError(error_message)
        else:
            pass
        
        return miclist
    

    def perform_splitscan_by_finding_location_with_respect_to_reference_micrograph(self, micrograph_files, outfiles):
        self.log.plog(10)
        for each_micrograph_index, each_micrograph_file in enumerate(micrograph_files):
            self.img, self.nx, self.ny = Micrograph().readmic(each_micrograph_file)
        
            miclist, decimation_factor_find = self.roughsplit(self.img, self.row_count, self.column_count,
            self.mic_count)
            
            self.log.plog(90 * (each_micrograph_index + 0.2) / len(micrograph_files) + 10)
            xoffset, yoffset, ccpeaks = self.findlabel(miclist, decimation_factor_find, self.scanstep)
            self.log.plog(90 * (each_micrograph_index + 0.5) / len(micrograph_files) + 10)
            
            miclist = self.finesplit(self.row_count, self.column_count, self.mwidth, self.mheight, self.scanstep,
            xoffset, yoffset, ccpeaks, self.cc_threshold, outfiles[each_micrograph_index])
            
            self.log.plog(90 * (each_micrograph_index + 0.8) / len(micrograph_files) + 10)
    
            if self.normscanoption: 
                self.normalized_mics = self.normscan(miclist)
            if self.binoption: 
                mics_binned = self.bin_micrographs(miclist, self.binfactor)
            if self.fprintoption: 
                self.finalprint(miclist)
            if self.cprintoption:
                self.contactprint(self.img, each_micrograph_file)
            self.log.plog(90 * (each_micrograph_index + 1) / len(micrograph_files) + 10)
            

    def perform_splitscan(self):
        if len(self.micrograph_files) < self.cpu_count:
            self.cpu_count = len(self.micrograph_files)
            self.feature_set.parameters['Number of CPUs']=self.cpu_count
        OpenMpi().setup_and_start_mpi_version_if_demanded(self.mpi_option, self.feature_set, self.cpu_count)
        self.tempdir = Temporary().mktmpdir(self.temppath)
        
        outfiles = Features().rename_series_of_output_files(self.micrograph_files, self.outfile)
        self.perform_splitscan_by_finding_location_with_respect_to_reference_micrograph(self.micrograph_files, outfiles)
        
        os.rmdir(self.tempdir)
        self.log.endlog(self.feature_set)
        
        
def main():
    # Option handling
    parset = ScanSplitPar()
    mergeparset = OptHandler(parset)

    ######## Program
    scan = ScanSplit(mergeparset)
    scan.perform_splitscan()
    
    
if __name__ == '__main__':
    main()



# Author: Carsten Sachse 17-Jan-2011
# Copyright: EMBL (2010 - 2018)
# License: see license.txt for details
"""
Spring2d - suite of programs that analyze two-dimensional images of helices
"""
from PyQt4.QtGui import QMainWindow, QWidget, QTabWidget, QGridLayout, QApplication
from spring.segment2d.segclassexam import SegClassExamPar
from spring.segment2d.segclasslayer import SegClassLayerPar
from spring.segment2d.seglayer2lattice import SegLayer2LatticePar
from spring.segment2d.segment import SegmentPar
from spring.segment2d.segmentalign2d import SegmentAlign2dPar
from spring.segment2d.segmentclass import SegmentClassPar
from spring.segment2d.segmentctfapply import SegmentCtfApplyPar
from spring.segment2d.segmentexam import SegmentExamPar
from spring.segment2d.segmentplot import SegmentPlotPar
from spring.segment2d.segmentselect import SegmentSelectPar
from spring.springgui.spring3d_launch import SpringGuiTop, Spring3dGuiScrollTabs
from spring.springgui.spring_menu import GuiActions
from textwrap import wrap
import os
import sys


class Segment2dGui(QWidget):

    def __init__(self, parent=None):
        QWidget.__init__(self)
        
        segment2dguitop = SpringGuiTop('spring2d_logo{0}png'.format(os.extsep))
        self.tabWidget = QTabWidget()

        self.segment_features = SegmentPar()
        self.segmentTab = Spring3dGuiScrollTabs(self.segment_features)

#         self.sctfapply_feature_set = SegmentCtfApplyPar()
#         self.segmentCtfApplyTab = Spring3dGuiScrollTabs(self.sctfapply_feature_set)

        self.segmentexam_features = SegmentExamPar()
        self.segmentExamTab = Spring3dGuiScrollTabs(self.segmentexam_features)

        self.segmentclass_features = SegmentClassPar()
        self.SegmentClassTab = Spring3dGuiScrollTabs(self.segmentclass_features)

        self.segmentalign2d_features = SegmentAlign2dPar()
        self.segmentAlign2dTab = Spring3dGuiScrollTabs(self.segmentalign2d_features)

        self.segclassexam_features = SegClassExamPar()
        self.segClassExamTab = Spring3dGuiScrollTabs(self.segclassexam_features)

        self.segclasslayer_features = SegClassLayerPar()
        self.segClassLayerTab = Spring3dGuiScrollTabs(self.segclasslayer_features)

        self.seglayer2lattice_features = SegLayer2LatticePar()
        self.segLayer2LatticeTab = Spring3dGuiScrollTabs(self.seglayer2lattice_features)
        
        self.segmentselect_features = SegmentSelectPar()
        self.segmentselectTab = Spring3dGuiScrollTabs(self.segmentselect_features)
        
        self.segmentplot_features = SegmentPlotPar()
        self.segmentPlotTab = Spring3dGuiScrollTabs(self.segmentplot_features)
        
        self.addSpring2dTabs()

        layout = QGridLayout()
        layout.addWidget(segment2dguitop.pic, 0, 0, 4, 1)
        layout.addWidget(self.tabWidget, 0, 1, 7, 7)
        layout.setColumnStretch(0, 0)
        self.setLayout(layout)


    def addSpring2dTabs(self):
        self.tabWidget.addTab(self.segmentTab, self.segment_features.progname.title())
        self.tabWidget.setTabToolTip(0, '\n'.join(wrap(self.segment_features.progname.title() + ' - ' + \
        self.segment_features.proginfo)))
                                     
        self.tabWidget.addTab(self.segmentExamTab, self.segmentexam_features.progname.title())
        self.tabWidget.setTabToolTip(1, '\n'.join(wrap(self.segmentexam_features.progname.title() + ' - ' + \
        self.segmentexam_features.proginfo)))
        
        self.tabWidget.addTab(self.SegmentClassTab, self.segmentclass_features.progname.title())
        self.tabWidget.setTabToolTip(2, '\n'.join(wrap(self.segmentclass_features.progname.title() + ' - ' + \
        self.segmentclass_features.proginfo)))
        
        self.tabWidget.addTab(self.segmentAlign2dTab, self.segmentalign2d_features.progname.title())
        self.tabWidget.setTabToolTip(3, '\n'.join(wrap(self.segmentalign2d_features.progname.title() + ' - ' + \
        self.segmentalign2d_features.proginfo)))
        
        self.tabWidget.addTab(self.segClassExamTab, self.segclassexam_features.progname.title())
        self.tabWidget.setTabToolTip(4, '\n'.join(wrap(self.segclassexam_features.progname.title() + ' - ' + \
        self.segclassexam_features.proginfo)))
        
        self.tabWidget.addTab(self.segClassLayerTab, self.segclasslayer_features.progname.title())
        self.tabWidget.setTabToolTip(5, '\n'.join(wrap(self.segclasslayer_features.progname.title() + ' - ' + \
        self.segclasslayer_features.proginfo)))
        
        self.tabWidget.addTab(self.segLayer2LatticeTab, self.seglayer2lattice_features.progname.title())
        self.tabWidget.setTabToolTip(6, '\n'.join(wrap(self.seglayer2lattice_features.progname.title() + ' - ' + \
        self.seglayer2lattice_features.proginfo)))
        
        self.tabWidget.addTab(self.segmentPlotTab, self.segmentplot_features.progname.title())
        self.tabWidget.setTabToolTip(7, '\n'.join(wrap(self.segmentplot_features.progname.title() + ' - ' + \
        self.segmentplot_features.proginfo)))
        

class Segment2dMain(QMainWindow, GuiActions):
    def __init__(self, parent=None):
        QMainWindow.__init__(self, parent)

        self.package = 'emspring'        
        TwodExamCentral = Segment2dGui()
        self.setCentralWidget(TwodExamCentral)

        # inherit body of layout of actions
        self.setupUi(self)

def main():
    app = QApplication(sys.argv)
    window = Segment2dMain()

    window.show()
    app.exec_()

if __name__ == '__main__':
    main()



# Author: Carsten Sachse
# Copyright: EMBL (2010 - 2018)
# License: see license.txt for details
"""
Program to classify windowed particles using SPARX's k-means clustering
"""
from spring.csinfrastr.csfeatures import Features
from spring.csinfrastr.csreadinput import OptHandler
from spring.particle2d.particlealign2d import ParticleAlign2d
from spring.segment2d.segmentclass import SegmentClass, SegmentClassPar
from spring.segment2d.segmentexam import SegmentExam
import numpy as np

class ParticleClassPar(SegmentClassPar):
    """
    Class to initiate default dictionary with input parameters including help and range values and 
    status dictionary
    """
    def __init__(self):
        # package/program identity
        self.package = 'emspring'
        self.progname = 'particleclass'
        self.proginfo = __doc__
        self.code_files = [self.progname]

        self.class_features = Features()
        self.feature_set = self.class_features.setup(self)
        
        self.define_parameters_and_their_properties()
        self.define_program_states()
    

    def add_mask_dimension_to_features(self):
        self.feature_set = self.class_features.set_particle_inner_and_outer_diameter(self.feature_set)
        

class ParticleClassPreparation(SegmentClass):
    def define_helix_or_particle_dimensions(self):
        self.inner_diameter, self.outer_diameter = self.p['Estimated inner and outer particle diameter in Angstrom']
        self.outer_diameter_in_pixel_ori = int(round(self.outer_diameter / self.pixelsize_ori))
        self.inner_diameter_in_pixel_ori = int(round(self.inner_diameter/ self.pixelsize_ori))
        
    def apply_binfactor_if_required(self):
        if self.binfactor > 1 and self.binoption is True:
            self.infilestack, self.image_dimension, self.outer_diameter_in_pixel, self.pixelsize = \
            SegmentExam().apply_binfactor(self.binfactor, self.infilestack, self.image_dimension_ori,
            self.outer_diameter_in_pixel_ori, self.pixelsize_ori)
            
            self.inner_diameter_in_pixel = self.inner_diameter / self.pixelsize
        else:
            self.outer_diameter_in_pixel = self.outer_diameter_in_pixel_ori
            self.pixelsize = self.pixelsize_ori
            self.image_dimension = self.image_dimension_ori

        self.maskwidthpix = self.outer_diameter_in_pixel
        self.maskwidthpix_ori = self.outer_diameter_in_pixel_ori
        
        
    def prepare_mask_ori(self):
        self.mask_ori = ParticleAlign2d().make_smooth_circular_mask(self.outer_diameter_in_pixel_ori,
        self.inner_diameter_in_pixel_ori, self.image_dimension_ori)
        
        return self.mask_ori
    
    
    def prepare_mask(self):
        self.mask = ParticleAlign2d().make_smooth_circular_mask(self.outer_diameter_in_pixel,
        self.inner_diameter_in_pixel, self.image_dimension)
        
        return self.mask
    

class ParticleClass(ParticleClassPreparation):
    def determine_minimal_segment_size(self):
        min_segsize = int(np.sqrt(self.outer_diameter_in_pixel ** 2 + self.outer_diameter_in_pixel ** 2) + 0.5)
        return min_segsize
    
    def add_mask_dimensions_to_align2d_features(self):
        self.aligndict['Estimated inner and outer particle diameter in Angstrom'] = ((self.inner_diameter,
        self.outer_diameter))
        
        program_to_be_launched = 'particlealign2d'
        
        return program_to_be_launched

def main():
    # Option handling
    parset = ParticleClassPar()
    mergeparset = OptHandler(parset)

    ######## Program
    stack = ParticleClass(mergeparset)
    stack.classify()

if __name__ == '__main__':
    main()



# Author: Carsten Sachse 08-Jun-2011
# Copyright: EMBL (2010 - 2018)
# License: see license.txt for details
from EMAN2 import EMData, EMUtil, Util
from collections import namedtuple
import os
import shutil
from spring.csinfrastr.csdatabase import SpringDataBase, base, SegmentTable, refine_base, RefinementCycleTable, \
    RefinementCycleSegmentTable, HelixTable, CtfMicrographTable
from spring.csinfrastr.csproductivity import OpenMpi, Temporary, Support, DiagnosticPlot
from spring.csinfrastr.csreadinput import OptHandler
from spring.segment2d.segment import Segment
from spring.segment2d.segmentexam import SegmentExam
from spring.segment2d.segmentselect import SegmentSelect
from spring.segment3d.refine.sr3d_diagnostics import SegmentRefine3dSummary
from spring.segment3d.refine.sr3d_parameters import SegmentRefine3dPar
from spring.segment3d.segclassreconstruct import SegClassReconstruct

from sqlalchemy.sql.expression import desc, asc, and_
from tabulate import tabulate

import numpy as np


class SegmentRefine3dCheck(SegmentRefine3dSummary):

    def remove_intermediate_files_if_desired(self, *files):
        if not self.keep_intermediate_files:
            for each_file in files:
                os.remove(each_file)


    def get_aim_dict(self):
        """
        >>> from spring.segment3d.refine.sr3d_main import SegmentRefine3d
        >>> SegmentRefine3d().get_aim_dict()
        {'high': 2, 'max': 3, 'medium': 1, 'low': 0}
        """
        return dict(zip(['low', 'medium', 'high', 'max'], list(range(4))))


    def check_whether_search_restraints_are_correctly_set_for_half_set_refinement(self, info_series):
        if self.halfset_refinement:
            aim_ref = self.get_aim_dict()
            half_info_ids = [each_info_id for each_info_id, each_info in enumerate(info_series) \
                          if aim_ref[each_info.resolution_aim] >= aim_ref[self.halfset_start]]
            
            for each_half_id in half_info_ids:
                if info_series[each_half_id].azimuthal_restraint == 180.0:
                    info_series[each_half_id] = info_series[each_half_id]._replace(azimuthal_restraint = 179.9999)
                if info_series[each_half_id].out_of_plane_restraint == 180.0:
                    info_series[each_half_id] = info_series[each_half_id]._replace(out_of_plane_restraint = 179.9999)
        
        return info_series
    

    def check_whether_search_restraints_and_refinement_option_are_correctly_set(self, info_series):
        if info_series[0].azimuthal_restraint < 180.0 and not self.resume_refinement_option or \
        info_series[0].out_of_plane_restraint < 180.0 and not self.resume_refinement_option:
            msg = 'You have specified a refinement strategy with a resolution aim of ' + \
            '\'{0}\' and an azimuthal/out-of-plane search restraint '.format(info_series[0].resolution_aim) + \
            '< 180.0 degrees. These parameters require a starting search angle that can be specified by ' + \
            '\'Continue refinement option\' and the corresponding \'refinement.db file\'. ' + \
            'Re-launch {0} with these options or release the '.format(self.feature_set.progname) + \
            'azimuthal/out-of-plane search restraint to 180 degrees.'

            raise ValueError(msg)


    def check_wether_search_restraints_make_sense_in_multi_model_refinement_and_half_set_refinement(self, info_series):
        aim_ref = self.get_aim_dict()
        if self.halfset_refinement and aim_ref[info_series[0].resolution_aim] >= aim_ref[self.halfset_start] and \
        len(self.helical_symmetries) > 1 and not self.resume_refinement_option:
                msg = 'You have specified \'independent half set refinement\' for your first resolution aim and ' + \
                'requested a multi-model refinement. This combination will only work if you have starting search ' + \
                'parameters, i.e. run the multi-model refinement at low resolution without independent ' + \
                'half set refinement and continue refinement with the independent half set refwinement option.'
                
                raise ValueError(msg)


    def estimate_required_tmp_disk_space(self):

        alignment_size = self.compute_alignment_size_in_pixels(self.alignment_size_in_A, self.ori_pixelsize)
        helixwidthpix = int(round(self.helixwidth / self.ori_pixelsize))
        prj_size = self.compute_prj_size_from_max_out_of_plane_tilt_and_diameter(alignment_size,
        self.out_of_plane_tilt_angle_range, helixwidthpix)
        
        if not self.high_resolution_aim and not self.max_resolution_aim:
            prj_count = self.azimuthal_angle_count * self.out_of_plane_tilt_angle_count
        else:
            prj_count = 25 * self.azimuthal_angle_count * self.out_of_plane_tilt_angle_count
            
        info_series, self.iteration_count = self.define_series_of_search_steps(self.ori_pixelsize, self.refine_strategy,
        (self.low_resolution_aim, self.low_resolution_ang_range, self.low_resolution_trans_range),
        (self.medium_resolution_aim, self.medium_resolution_ang_range, self.medium_resolution_trans_range),
        (self.high_resolution_aim, self.high_resolution_ang_range, self.high_resolution_trans_range),
        (self.max_resolution_aim, self.max_resolution_ang_range, self.max_resolution_trans_range), self.iteration_count)
        
        info_series = self.check_whether_search_restraints_are_correctly_set_for_half_set_refinement(info_series)

        self.check_whether_search_restraints_and_refinement_option_are_correctly_set(info_series)       
        self.check_wether_search_restraints_make_sense_in_multi_model_refinement_and_half_set_refinement(info_series)

        min_binfactor = min([each_info.bin_factor for each_info in info_series])
        
        model_count = len(self.helical_symmetries)

        byte_size = model_count * int(Support().compute_byte_size_of_image_stack(prj_size, prj_size, prj_count) /
        float(min_binfactor))
        
        return byte_size, info_series
    

    def check_helical_rise_and_segmentation_step(self, helical_symmetry):
        helical_rise, helical_rotation = helical_symmetry
        if helical_rise == 0 and helical_rotation == 0:
            sym_view_count = 1
        else:
            sym_view_count = SegClassReconstruct().determine_symmetry_view_count(
                (self.alignment_size_in_A + self.stepsize) / self.ori_pixelsize, self.alignment_size_in_A / self.ori_pixelsize, 
                (helical_rise, helical_rotation), self.ori_pixelsize, 1)
            if helical_rise > 1.4 * self.stepsize:
                msg = 'Helical rise of {0} Angstrom is significantly larger than '.format(helical_rise)
                msg += 'segmentation step size of {0} Angstrom. '.format(self.stepsize)
                msg += 'This will result in multiple inclusion of the same asymmetric unit '
                msg += 'in the 3D final reconstruction. '
                msg += 'Please run \'segment\' again with a step size >= {0} '.format(helical_rise)
                msg += 'Angstrom (preferably using a multiple of the helical rise, e.g. '
                msg += '{0}, {1}, {2}...))'.format(helical_rise, 2 * helical_rise, 3 * helical_rise)
                raise ValueError(msg)
            
        return sym_view_count


    def copy_spring_db_to_tempdir(self):
        temp_db = os.path.join(self.tempdir, 'spring_temp.db')
        shutil.copy('spring.db', temp_db)

        return temp_db


    def get_frame_id_from_mic_stack_file(self, segment_mic):
        """
        >>> from spring.segment3d.refine.sr3d_main import SegmentRefine3d
        >>> s = SegmentRefine3d()
        >>> from collections import namedtuple
        >>> mic = namedtuple('mic', 'micrograph_name')
        >>> segment_mic = mic('stack_name@2.mrcs')
        >>> s.get_frame_id_from_mic_stack_file(segment_mic)
        2
        """
        return int(segment_mic.micrograph_name.split('@')[-1].strip(os.extsep + 'mrcs'))


    def determine_upper_and_lower_bound_from_central_frame_id(self, frame_avg_window, frame_id, frame_count):
        """
        >>> from spring.segment3d.refine.sr3d_main import SegmentRefine3d
        >>> s = SegmentRefine3d()
        >>> from collections import namedtuple
        >>> mic = namedtuple('mic', 'micrograph_name')
        >>> segment_mic = mic('stack_name@2.mrcs')
        >>> s.determine_upper_and_lower_bound_from_central_frame_id(3, 4, 7)
        (3, 5)
        >>> s.determine_upper_and_lower_bound_from_central_frame_id(3, 0, 7)
        (0, 2)
        >>> s.determine_upper_and_lower_bound_from_central_frame_id(5, 1, 7)
        (0, 4)
        >>> s.determine_upper_and_lower_bound_from_central_frame_id(5, 6, 7)
        (2, 6)
        """
        lower_frame_id = frame_id - (frame_avg_window - 1) / 2
        upper_frame_id = frame_id + (frame_avg_window - 1) / 2
        
        if lower_frame_id < 0:
            diff = 0 - lower_frame_id 
            lower_frame_id += diff
            upper_frame_id += diff
        elif upper_frame_id >= frame_count:
            diff = frame_count - 1 - upper_frame_id
            lower_frame_id += diff
            upper_frame_id += diff

        return lower_frame_id, upper_frame_id


    def assert_segment_count_on_stack_equals_that_in_database(self):
        segment_count = EMUtil.get_image_count(self.infile)
        shutil.copy(self.spring_path, 'spring.db')
        temp_db = self.copy_spring_db_to_tempdir()
        session = SpringDataBase().setup_sqlite_db(base, temp_db)
        
        database_count = session.query(SegmentTable).count()
        session.close()

        if segment_count != database_count:
            msg = 'The specified spring.db file does not have the same number of segments than the provided segment '
            msg += 'stack. Make sure that the correct input segment stack and the corresponding spring.db generated '
            msg += 'by the same segmentation run are specified.'
            raise ValueError(msg)
        
        self.curvature_range, self.ccc_layer_range = SegmentSelect().convert_curvature_ccc_layer_range('spring.db',
        self.straightness_selection, self.curvature_range_perc, self.ccc_layer_selection, self.ccc_layer_range_perc)
        
        image_list, excluded_non_orientation_counts = \
        SegmentSelect().filter_non_orientation_parameters_based_on_selection_criteria(self, temp_db,
        keep_helices_together=True)
        
        if self.frame_motion_corr and self.frame_avg_window > 2:
            image_frame_list = []
            first_mic = session.query(CtfMicrographTable).first()
            if '@' not in first_mic.micrograph_name:
                err_msg = 'Your spring.db does not the required information about frames. Re-create the appropriate ' + \
                'frame information by running \'segment\' with the \'Frame processing option\'.'
                raise ValueError(err_msg)
            
            frame_counts = []
            for each_image_id in image_list:
                segment = session.query(SegmentTable).filter(SegmentTable.stack_id == each_image_id).first()
                segment_mic = session.query(CtfMicrographTable).filter(CtfMicrographTable.id == segment.mic_id).first()

                frame_prefix = segment_mic.micrograph_name.split('@')[0]
                frame_id = self.get_frame_id_from_mic_stack_file(segment_mic)

                frame_mics = session.query(CtfMicrographTable).\
                filter(CtfMicrographTable.micrograph_name.startswith(frame_prefix)).all()
                
                frame_count = len([each_frame_mic for each_frame_mic in frame_mics])
                frame_counts.append(frame_count)

                lower_frame_id, upper_frame_id = \
                self.determine_upper_and_lower_bound_from_central_frame_id(self.frame_avg_window, frame_id, frame_count)

                frame_mic_ids = [each_frame_mic.id for each_frame_mic in frame_mics \
                if lower_frame_id <= self.get_frame_id_from_mic_stack_file(each_frame_mic) <= upper_frame_id]

                frame_segments = session.query(SegmentTable).\
                filter(SegmentTable.picked_x_coordinate_A == segment.picked_x_coordinate_A).\
                filter(SegmentTable.mic_id.in_(frame_mic_ids)).all()

                frame_segment_ids = [each_image_id] + \
                                    [each_frame_segment.stack_id for each_frame_segment in frame_segments \
                                     if each_frame_segment.stack_id != each_image_id] 

                image_frame_list.append([each_image_id, frame_segment_ids])

            image_list = list(image_frame_list)
            if len(set(frame_counts)) > 0:
                self.frame_count = list(set(frame_counts))[0]
            else:
                err_msg = 'You have provided a spring.db database file that contains different number of frames ' + \
                'for different segments. Please make sure the spring.db was generated with \'segment\' using the ' + \
                '\'frame averaging option\'.'
                raise ValueError(msg)

        os.remove(temp_db)
        
        return segment_count, image_list


    def get_last_cycle_from_refinement_database(self, refine_db):
        ref_session = SpringDataBase().setup_sqlite_db(refine_base, refine_db)
        last_cycle = ref_session.query(RefinementCycleTable).order_by(desc(RefinementCycleTable.id)).first()
        
        return last_cycle, ref_session
    

    def get_original_image_list(self, image_list):
        if self.frame_motion_corr and self.frame_avg_window > 2:
            ori_image_list, b = zip(*image_list)
        else:
            ori_image_list = list(image_list)

        return ori_image_list


    def check_whether_ref_db_has_fewer_segment_entries_than_requested_in_case_of_continued_refinement(self, image_list):
        if self.resume_refinement_option:
            shutil.copy(self.refinementdb_path, 'refinement.db')
            temp_ref_db = self.copy_ref_db_to_tempdir()
            last_cycle, ref_session = self.get_last_cycle_from_refinement_database(temp_ref_db)

            orientation_results = ref_session.query(RefinementCycleSegmentTable).\
            filter(RefinementCycleSegmentTable.cycle_id == last_cycle.id).\
            order_by(asc(RefinementCycleSegmentTable.stack_id)).all()

            ref_session.close()
            os.remove(temp_ref_db)
        
            ref_stack_ids = set([each_ref_segment.stack_id for each_ref_segment in orientation_results])
            ori_image_list = self.get_original_image_list(image_list)

            if len(ref_stack_ids.intersection(ori_image_list)) < len(ori_image_list):
                msg = 'This refinement cannot be continued as requested. Your previous refinement was run with ' + \
                'less segments than the currently requested one. This is probably due to a more stringent segment ' + \
                'selection this time. If you want to work with this segment selection re-launch the entire ' + \
                'refinement from the start instead.'  
                raise ValueError(msg)


    def check_entry_count_for_segmultirefine3d(self):
        max_entries = max(len(self.helical_symmetries), len(self.polar_helices), len(self.rotational_symmetry_starts))
        min_entries = min(len(self.helical_symmetries), len(self.polar_helices), len(self.rotational_symmetry_starts))
        
        if max_entries != min_entries:
            msg = 'You have not entered the same number of values for \'Helical symmetries\', \'Polar helices\' and ' + \
            '\'Rotational symmetries\'. Instead {0} received: '.format(self.feature_set.progname) + \
            'Helical symmetries: {0} (counted: {1}). '.format(
            ', '.join([each_sym.__str__() for each_sym in self.helical_symmetries]), len(self.helical_symmetries)) + \
            'Polar helices: {0} (counted: {1}). '.format(', '.join(self.polar_helices), len(self.polar_helices)) + \
            'Rotational symmetries: {0} (counted: {1}).'.format(
            ', '.join([each_rot.__str__() for each_rot in self.rotational_symmetry_starts]), len(self.rotational_symmetry_starts))

            raise ValueError(msg)

        if len(self.references) > 1 and len(self.references) != max_entries and self.reference_option:
            msg = 'You have not entered the same number of references as for other entries that are required ' + \
            'in competitive refinement in {0}. '.format(self.feature_set.progname)
            'Helical symmetries: {0} (counted: {1}). '.format(
            ', '.join([each_sym.__str__() for each_sym in self.helical_symmetries]), len(self.helical_symmetries)) + \
            'References: {0} (counted: {1})'.format(
            ', '.join([os.path.basename(each_ref) for each_ref in self.references]), len(self.references))
            
            raise ValueError(msg)

            
    def assert_at_least_view_count_of_100_for_3dreconstruction(self, sym_view_count, polar_helix,
    rotational_symmetry_start, segment_count):
        sym_transformations = SegClassReconstruct().get_symmetry_transformations_from_helix_input(polar_helix,
        rotational_symmetry_start)
        
        rec_view_count = len(sym_transformations) * sym_view_count * int(segment_count / float(self.cpu_count))
        if rec_view_count < 100 and self.mpi_option:
            msg = '{0} CPUs requested for {1} images in stack. '.format(self.cpu_count, segment_count)
            msg += 'With the given symmetry-related input this will result in {0} images '.format(rec_view_count)
            msg += '(less than 100) per parallel reconstruction. '
            msg += 'Please lower number of requested CPUs, e.g. Number of CPUs {0} '.format(int(
                    self.cpu_count * rec_view_count / 100.0))
            msg += '= (if number of images for reconstruction is 100).'
            raise ValueError(msg)
        

    def check_sanity_of_input_parameters(self):
        segment_count, image_list = self.assert_segment_count_on_stack_equals_that_in_database() 

        self.check_whether_ref_db_has_fewer_segment_entries_than_requested_in_case_of_continued_refinement(image_list)
            
        self.check_entry_count_for_segmultirefine3d()

        for each_helical_sym, each_point_sym, each_rot_start in zip(self.helical_symmetries, self.polar_helices,
        self.rotational_symmetry_starts):
            sym_view_count = self.check_helical_rise_and_segmentation_step(each_helical_sym)

            self.assert_at_least_view_count_of_100_for_3dreconstruction(sym_view_count, each_point_sym, each_rot_start,
            segment_count)
        
        return image_list
        
            
    def copy_ref_db_to_tempdir(self, cycle_id=None):
        if cycle_id is None:
            temp_ref_db = os.path.join(self.tempdir, 'ref_temp.db')
            shutil.copy('refinement.db', temp_ref_db)
        else:
            temp_ref_db = os.path.join(self.tempdir, 'ref_temp_{0:03}.db'.format(cycle_id))
            shutil.copy('refinement{0:03}.db'.format(cycle_id), temp_ref_db)

        return temp_ref_db


    def merge_model_ids_into_presplit_state(self, ref_session, last_cycle, reference_files):

        ref_segments = ref_session.query(RefinementCycleSegmentTable).all()
        model_ids = set([each_ref_segment.model_id for each_ref_segment in ref_segments])
        if len(model_ids) != len(reference_files) * 2:
            msg = 'You have specified to continue a refinement with a refinement.db that has been used for ' + \
            'independent half-set refinement. The number of models you specified to refine does not match the ' + \
            'number from the half-set refinement. Please adapt your input or activate half-set refinement.'
            raise ValueError(msg)
        
        for each_reference in reference_files: 
            ref_segments = ref_session.query(RefinementCycleSegmentTable).\
            filter(RefinementCycleSegmentTable.cycle_id == last_cycle.id).\
            filter(and_(RefinementCycleSegmentTable.model_id >= 2 * each_reference.model_id, 
                        RefinementCycleSegmentTable.model_id <= 2 * each_reference.model_id + 1)).all()
            
            for each_ref_segment in ref_segments:
                each_ref_segment.model_id = each_reference.model_id
                ref_session.merge(each_ref_segment)

        last_cycle.fsc_split = False
        ref_session.merge(last_cycle)
        ref_session.commit()

        return ref_session
    

    def get_last_cycle_id_from_refinement_database(self, refinementdb_path, reference_files):
        shutil.copy(self.refinementdb_path, 'refinement.db')
        temp_ref_db = self.copy_ref_db_to_tempdir()
        last_cycle, ref_session = self.get_last_cycle_from_refinement_database(temp_ref_db)

        if last_cycle is None:
            ref_cycle_id = 0
            ref_session.close()
            os.remove(temp_ref_db)
        else:
            ref_cycle_id = last_cycle.id
            if last_cycle.fsc_split and not self.halfset_refinement:
                ref_session = self.merge_model_ids_into_presplit_state(ref_session, last_cycle, reference_files)
                ref_session.close()
            shutil.move(temp_ref_db, 'refinement{0:03}.db'.format(ref_cycle_id))
            os.remove('refinement.db')
            
        return ref_cycle_id
    

    def update_image_list_after_file_transfer(self, ori_image_list, image_list):
        """
        >>> from spring.segment3d.refine.sr3d_main import SegmentRefine3d
        >>> s = SegmentRefine3d()
        >>> s.frame_motion_corr = False
        >>> s.frame_avg_window = 5
        >>> ori_image_list = list(range(0, 10, 2))
        >>> s.update_image_list_after_file_transfer(ori_image_list, ori_image_list)
        [0, 1, 2, 3, 4]
        >>> s.frame_motion_corr = True
        >>> frame_list = [[0, 2, 4], [0, 2, 4], [2, 4, 6], [4, 6, 8], [4, 6, 8]]
        >>> image_list = zip(ori_image_list, frame_list)
        >>> s.update_image_list_after_file_transfer(ori_image_list, image_list)
        [(0, [0, 1, 2]), (1, [0, 1, 2]), (2, [1, 2, 3]), (3, [2, 3, 4]), (4, [2, 3, 4])]
        """
        if self.frame_motion_corr and self.frame_avg_window > 2:
            img_ids, frame_img_ids = zip(*image_list)
            updated_frame_img_ids = []
            for each_local_id, each_image_number in enumerate(ori_image_list):
                updated_frame_ids = [each_local_id_id for each_local_id_id, each_image_id in enumerate(ori_image_list) \
                                         if each_image_id in frame_img_ids[each_local_id]]

                updated_frame_img_ids.append(updated_frame_ids)
            
            image_list = zip(list(range(len(ori_image_list))), updated_frame_img_ids)
        else:
            image_list = list(range(len(ori_image_list)))

        return image_list


    def pre_cycle_setup(self, info_series):
        image_list = self.check_sanity_of_input_parameters()
        reference_files = self.prepare_reference_volumes()
        
        masked_segment_stack = '{inp}_masked{ext}'.format(inp=os.path.splitext(os.path.basename(self.infile))[0],
        ext=os.path.splitext(self.infile)[-1])
        
        large_segment_stack = os.path.splitext(os.path.basename(self.infile))[0] + '_temp' + \
        os.path.splitext(os.path.basename(self.infile))[-1]
        
        ori_image_list = self.get_original_image_list(image_list)
        segment = EMData()
        for each_local_id, each_image_number in enumerate(ori_image_list):
            segment.read_image(self.infile, each_image_number)
            segment.write_image(large_segment_stack, each_local_id)
            
        image_list = self.update_image_list_after_file_transfer(ori_image_list, image_list)

        self.tempfiles = []
        self.tempfiles.append(large_segment_stack)
            
        if self.resume_refinement_option:
            ref_cycle_id = self.get_last_cycle_id_from_refinement_database(self.refinementdb_path, reference_files)
        else:
            ref_cycle_id = 0

        lambda_sirt = 1.0e-5
        
        return large_segment_stack, masked_segment_stack, image_list, reference_files, lambda_sirt, ref_cycle_id
    

    def define_original_input_values_before_binning(self, large_segment_stack, reference_files):
        ori_large_segment_stack = large_segment_stack
        ori_reference_files = reference_files
        alignment_size = self.compute_alignment_size_in_pixels(self.alignment_size_in_A, self.ori_pixelsize)
        ori_alignment_size = alignment_size
        ori_pixelsize = self.ori_pixelsize
        ori_helixwidthpix = int(round(self.helixwidth / self.ori_pixelsize))
        
        return ori_large_segment_stack, ori_reference_files, ori_alignment_size, ori_helixwidthpix, ori_pixelsize
    
    
    def volume_decimate(self, vol, decimation=2, fit_to_fft = True, frequency_low=0, frequency_high=0):
        """
            Window 3D volume to FFT-friendly size, apply Butterworth low pass filter,
            and decimate image by integer factor
            cs: adapted from image_decimate (fundamentals)
        """
        from filter       import filt_btwl
        from fundamentals import smallprime
        
        if decimation == 1.0:     
            return  vol.copy()
        if frequency_low <= 0  :    
            frequency_low = 0.5/decimation-0.02
            frequency_high = min(0.5/decimation + 0.02, 0.499)
        if fit_to_fft:
            nx = vol.get_xsize()
            ny = vol.get_ysize()
            nz = vol.get_ysize()
            nx_fft_m = smallprime(nx)
            ny_fft_m = smallprime(ny)
            nz_fft_m = smallprime(nz)
            e = Util.window(vol, nx_fft_m, ny_fft_m, nz_fft_m, 0,0,0)
            e = filt_btwl(e, frequency_low, frequency_high)
        else:
            e = filt_btwl(vol, frequency_low, frequency_high)
        
        return Util.decimate(e, int(decimation), int(decimation), int(decimation))


    def give_name_and_write_out_reference(self, reference_vol, each_reference, ref_cycle_id, updated_references):
        ref_prefix = '{0}_rescale_mod{1:03}_{2:03}{3}'.format(os.path.splitext(self.outfile_prefix)[0],
        each_reference.model_id, ref_cycle_id, os.path.splitext(self.outfile_prefix)[-1])

        new_reference = self.generate_reconstruction_name_with_Angstrom_per_pixel_and_set_header(0, reference_vol, 
        ref_prefix, self.ori_pixelsize, each_reference.helical_symmetry, each_reference.point_symmetry)

        reference_vol.write_image(new_reference)
        each_reference = each_reference._replace(ref_file=new_reference)
        updated_references.append(each_reference)
        
        return updated_references


    def make_pixel_info_named_tuple(self):
        return namedtuple('pixel_info', 'pixelsize alignment_size reconstruction_size helixwidthpix ' + \
        'helix_inner_widthpix helix_heightpix')


    def update_reference_info(self, reference_files):
        upd_reference_files = []
        for each_model_id, each_reference in enumerate(reference_files):
            even_model_id = each_model_id * 2
            each_first_reference = each_reference._replace(model_id=even_model_id)
            each_second_reference = each_reference._replace(model_id=even_model_id + 1)
            upd_reference_files.append(each_first_reference)
            upd_reference_files.append(each_second_reference)
        
        return upd_reference_files


    def distribute_segment_ids_into_two_halves_based_on_helices(self, stack_ids, helix_ids, length_sorted_helices):
        """
        >>> from spring.segment3d.refine.sr3d_main import SegmentRefine3d
        >>> s = SegmentRefine3d()
        >>> s.distribute_segment_ids_into_two_halves_based_on_helices(range(5), [0] * 2 + [1] * 3, [0, 1])
        (array([0, 1]), array([2, 3, 4]))
        >>> s.distribute_segment_ids_into_two_halves_based_on_helices(range(5), [0] * 5, [0])
        (array([0, 1]), array([2, 3, 4]))
        """
        even_helices = [each_helix_id for each_helix, each_helix_id in enumerate(length_sorted_helices) \
                        if (each_helix) % 2 == 0]

        even_stack_ids = np.array([each_stack_id for (each_stack_id, each_helix_id) in zip(stack_ids, helix_ids) \
                                   if each_helix_id in even_helices])

        odd_stack_ids = np.array([each_stack_id for (each_stack_id, each_helix_id) in zip(stack_ids, helix_ids) \
                                  if each_helix_id not in even_helices])
        
        if len(even_stack_ids) == 0 or len(odd_stack_ids) == 0:
            even_stack_ids = np.array(stack_ids[:len(stack_ids) / 2])
            odd_stack_ids = np.array(stack_ids[len(stack_ids) / 2:])

        return even_stack_ids, odd_stack_ids


    def split_ref_ids_according_to_even_and_odd_helices(self, reference_files, session, ref_session, last_cycle):
        model_ref_segments = []
        for each_reference in reference_files:
            model_segments = ref_session.query(RefinementCycleSegmentTable).\
            filter(RefinementCycleSegmentTable.cycle_id == last_cycle.id).\
            filter(RefinementCycleSegmentTable.model_id == each_reference.model_id).\
            order_by(asc(RefinementCycleSegmentTable.stack_id)).all()

            ref_segment_ids = np.array([each_ref_segment.id for each_ref_segment in model_segments])
            stack_ids = [each_ref_segment.stack_id for each_ref_segment in model_segments]
            helix_ids = []
            for each_stack_id in stack_ids:
                segment = session.query(SegmentTable).get(each_stack_id + 1)
                helix_ids.append(segment.helix_id)
            
            helices = session.query(HelixTable).order_by(desc(HelixTable.length)).all()
            length_sorted_helices = [each_helix.id for each_helix in helices]

            even_stack_ids, odd_stack_ids = self.distribute_segment_ids_into_two_halves_based_on_helices(stack_ids,
            helix_ids, length_sorted_helices)

            model_ref_segments.append(ref_segment_ids[np.in1d(stack_ids, odd_stack_ids)])
            model_ref_segments.append(ref_segment_ids[np.in1d(stack_ids, even_stack_ids)])
        
        return model_ref_segments


    def prepare_reference_files_for_halfset_refinement(self, reference_files, cycle_id):
        
        self.fsc_split = True
        temp_db = self.copy_spring_db_to_tempdir()
        session = SpringDataBase().setup_sqlite_db(base, temp_db)
        
        temp_ref_db = self.copy_ref_db_to_tempdir(cycle_id)
        ref_session = SpringDataBase().setup_sqlite_db(refine_base, temp_ref_db)

        upd_reference_files = self.update_reference_info(reference_files)

        last_cycle = ref_session.query(RefinementCycleTable).order_by(desc(RefinementCycleTable.id)).first()

        if not last_cycle.fsc_split:
            model_ref_segments = self.split_ref_ids_according_to_even_and_odd_helices(reference_files, session, ref_session,
            last_cycle)
            
            for each_model_id, each_model_ref_segments in enumerate(model_ref_segments):
                for each_model_id_segment in each_model_ref_segments:
                    ref_segment = ref_session.query(RefinementCycleSegmentTable).get(int(each_model_id_segment))
                    ref_segment.model_id = each_model_id
                    ref_session.merge(ref_segment)
    
            ref_session.commit()
            ref_session.close()
        
        os.remove(temp_db)
        shutil.copy(temp_ref_db, 'refinement{0:03}.db'.format(cycle_id))
        os.remove(temp_ref_db)

        return upd_reference_files


    def update_pixelinfo_based_on_different_helical_symmetries(self, segment_size_pix, helixwidthpix, pixelsize):
        alignment_sizes = []
        reconstruction_sizes = []
        helix_heights = []
        for each_helical_symmetry in self.helical_symmetries:
            helical_rise, helical_rotation = each_helical_symmetry

            alignment_size, reconstruction_size = \
            self.compute_alignment_and_reconstruction_size_in_pixels(self.alignment_size_in_A, helical_rise,
            self.helixwidth, pixelsize)

            helix_heights.append(min(alignment_size, int(self.alignment_size_in_A / pixelsize)))
            if helical_rise == 0 and helical_rotation == 0:
                reconstruction_size = alignment_size
            else:
                reconstruction_size = self.compute_rec_size_for_helix(self.helixwidth, helical_rise, pixelsize)

            reconstruction_sizes.append(min(reconstruction_size, segment_size_pix))
            alignment_sizes.append(min(alignment_size, segment_size_pix))
        
        pixelinfo_nt = self.make_pixel_info_named_tuple()

        helix_inner_widthpix = int(round(self.helix_inner_width / (pixelsize)))

        pixelinfo = pixelinfo_nt(pixelsize, max(alignment_sizes), max(reconstruction_sizes), helixwidthpix,
        helix_inner_widthpix, max(helix_heights))

        return pixelinfo, reconstruction_sizes


    def precycle_setup_before_binning(self, info_series, large_segment_stack, reference_files, ori_reference_files,
    ori_pixelsize, each_binindex, ref_cycle_id, segment_size_pix, helixwidthpix):
        reference_vol = EMData()
        current_binfactor = info_series[each_binindex].bin_factor
        current_res_aim = info_series[each_binindex].resolution_aim
        updated_references = []
        if current_res_aim == info_series[0].resolution_aim:
            aim_ref = self.get_aim_dict()
            if self.halfset_refinement and aim_ref[info_series[0].resolution_aim] >= aim_ref[self.halfset_start] and\
            self.reference_option:
                ori_reference_files = self.prepare_reference_files_for_halfset_refinement(ori_reference_files,
                ref_cycle_id)

            for each_ref_id, each_ori_ref in enumerate(ori_reference_files):
                reference_vol.read_image(each_ori_ref.ref_file)
                reference_vol = self.volume_decimate(reference_vol, current_binfactor, fit_to_fft=False)
                ref_size = reference_vol.get_xsize()
                reference_vol = Util.window(reference_vol, ref_size, ref_size, ref_size, 0, 0, 0)

                updated_references = self.give_name_and_write_out_reference(reference_vol, each_ori_ref, ref_cycle_id,
                updated_references)
                if self.halfset_refinement and aim_ref[info_series[0].resolution_aim] >= aim_ref[self.halfset_start]:
                    if each_ref_id % 2 == 1:
                        os.remove(each_ori_ref.ref_file)
                else:
                    os.remove(each_ori_ref.ref_file)
        else:
            if self.halfset_refinement and current_res_aim == self.halfset_start:
                reference_files = self.prepare_reference_files_for_halfset_refinement(reference_files,
                ref_cycle_id)

            for each_reference in reference_files:
                reference_vol.read_image(each_reference.ref_file)
                scaling_factor = info_series[each_binindex - 1].bin_factor / float(current_binfactor)
                ref_size = Segment().determine_boxsize_closest_to_fast_values(reference_vol.get_xsize() * scaling_factor)
                reference_vol = Util.pad(reference_vol, ref_size, ref_size, ref_size, 0, 0, 0, 'avg')
                reference_vol.scale(scaling_factor)
                
                updated_references = self.give_name_and_write_out_reference(reference_vol, each_reference, ref_cycle_id,
                updated_references)
       
        pixelsize = ori_pixelsize * current_binfactor

        pixelinfo, reconstruction_sizes = \
        self.update_pixelinfo_based_on_different_helical_symmetries(segment_size_pix, helixwidthpix,
        pixelsize)

        if self.ctf_correction:
            ctf3d_avg_squared = self.prepare_3dctf_avg_squared(max(reconstruction_sizes), pixelsize)
        else:
            ctf3d_avg_squared = None
        
        return updated_references, ctf3d_avg_squared, pixelinfo
    

    def generate_reconstruction_name_with_Angstrom_per_pixel_and_set_header(self, ref_cycle_id, reconstruction,
    outfile_prefix, pixelsize, helical_symmetry, point_symmetry):
    
        latest_reconstruction = self.generate_file_name_with_apix(ref_cycle_id, outfile_prefix, pixelsize)
        
        reconstruction = SegClassReconstruct().set_isotropic_pixelsize_in_volume(pixelsize, reconstruction)
        reconstruction = SegClassReconstruct().set_header_with_helical_parameters(helical_symmetry, reconstruction,
        point_symmetry)
        
        return latest_reconstruction
    
    
    def clean_up_temporary_large_stack(self, large_segment_stack, masked_segment_stack):
        temp_large_segment_stack = os.path.join(self.tempdir, os.path.basename(large_segment_stack))
        if os.path.exists(temp_large_segment_stack):
            os.remove(temp_large_segment_stack)
        os.remove(masked_segment_stack)


class SegmentRefine3d(SegmentRefine3dCheck):
    def setup_dummies_for_iteration(self, start_ref_cycle_id):
        unbending_info = None
        large_straightened_stack = None
        ref_cycle_id = start_ref_cycle_id
        
        return ref_cycle_id, unbending_info, large_straightened_stack
    

    def get_prj_named_tuple(self):
        prj_nt = namedtuple('prj_info', 'projection_stack fine_projection_stack projection_parameters ' + \
        'fine_projection_parameters')

        return prj_nt


    def package_parameters_and_stack_name_into_prj_info(self, merged_prj_params, merged_fine_prj_params,
    merged_prj_stack, merged_prj_fine_stack):
        prj_nt = self.get_prj_named_tuple()
        prj_info = prj_nt(merged_prj_stack, merged_prj_fine_stack, merged_prj_params, merged_fine_prj_params)

        return prj_info


    def project_including_masking_and_filtering(self, required_byte, info_series, reference_files, start_ref_cycle_id,
    ref_cycle_id, each_info, each_iteration_number, pixelinfo):
        
        Temporary().check_available_space_in_temppath_and_raise_error_if_not_enough_space_available(self.temppath,
        required_byte)
        
        updated_ref_files = []
        merged_prj_params = []
        merged_fine_prj_params = []
        if self.resume_refinement_option and not self.reference_option and each_iteration_number == 1:
            merged_prj_stack = None
            merged_prj_fine_stack = None
            updated_ref_files = reference_files
        else:
            for each_reference in reference_files:
                reference_volume = self.filter_and_mask_reference_volume(each_info.resolution_aim, each_reference,
                pixelinfo, each_reference.fsc)
            
                each_reference, prj_prefix = \
                self.write_out_reference_and_get_prj_prefix_depending_on_number_of_models(reference_files, ref_cycle_id,
                each_iteration_number, each_reference, reference_volume)
    
                projection_stack, projection_parameters, fine_projection_stack, fine_projection_parameters = \
                self.project_through_reference_volume_in_helical_perspectives(each_info.resolution_aim, ref_cycle_id,
                each_reference.ref_file, pixelinfo, each_reference.helical_symmetry, each_reference.rotational_symmetry,
                prj_prefix)
                
                updated_ref_files, merged_prj_params, merged_fine_prj_params = \
                self.collect_prj_params_and_update_reference_info(updated_ref_files, each_reference, projection_stack,
                projection_parameters, fine_projection_stack, fine_projection_parameters, merged_prj_params,
                merged_fine_prj_params)
        
            merged_prj_stack = self.merge_prj_ref_stacks_into_single_prj_stack(updated_ref_files, 'prj_stack')
            merged_prj_fine_stack = self.merge_prj_ref_stacks_into_single_prj_stack(updated_ref_files, 'fine_prj_stack')
        
        self.log.plog(90 * (ref_cycle_id - start_ref_cycle_id + 0.2 - 1) / float(self.iteration_count) + 5)

        previous_params, mask_params =\
        self.prepare_previous_parameters_either_from_inplane_angle_or_from_previous_cycle(each_info, info_series,
        ref_cycle_id, each_iteration_number, reference_files)
        
        prj_info = self.package_parameters_and_stack_name_into_prj_info(merged_prj_params, merged_fine_prj_params,
        merged_prj_stack, merged_prj_fine_stack)

        return mask_params, previous_params, prj_info, updated_ref_files
        

    def mask_and_window_and_unbend_if_required(self, info_series, masked_segment_stack, start_ref_cycle_id,
    ref_cycle_id, each_info, large_segment_stack, each_iteration_number, mask_params, previous_params, unbending_info,
    large_straightened_stack, pixelinfo):
        
        if not self.unbending and each_iteration_number == 1 or \
        not self.unbending and self.layer_line_filter:
            masked_segment_stack, large_segment_stack, large_straightened_stack = \
            self.window_and_mask_input_stack(large_segment_stack, pixelinfo, mask_params, masked_segment_stack,
            each_info, ref_cycle_id)
            
        elif self.unbending and each_iteration_number == 1:
            helices_coordinates, cut_coordinates = \
            self.get_helices_coordinates_required_for_unbending_from_database(ref_cycle_id - 1, each_info.bin_factor,
            info_series, large_segment_stack, pixelinfo.pixelsize)
            
            previous_params, unbending_info, masked_segment_stack, large_segment_stack, large_straightened_stack = \
            self.unbend_window_and_mask_input_stack(large_segment_stack, ref_cycle_id, pixelinfo,
            previous_params, mask_params, helices_coordinates, cut_coordinates, masked_segment_stack,
            each_info.resolution_aim)
            
        self.log.plog(90 * (ref_cycle_id - start_ref_cycle_id + 0.3 - 1) / float(self.iteration_count) + 5)

        return masked_segment_stack, previous_params, unbending_info, large_straightened_stack, large_segment_stack


    def add_model_id_to_prefix(self, prefix, model_id):
        outfile_prefix = os.path.splitext(prefix)[0] + '_mod{0:03}'.format(model_id) + os.path.splitext(prefix)[-1]

        return outfile_prefix


    def add_iter_id_to_prefix(self, prefix, ref_cycle_id):
        outfile_prefix = os.path.splitext(prefix)[0] + '_{0:03}'.format(ref_cycle_id) + os.path.splitext(prefix)[-1]

        return outfile_prefix


    def write_out_reconstruction_and_remove_reference(self, reference_files, ref_cycle_id, pixelinfo, each_reference,
    reconstructed_volume):
        if len(reference_files) > 1:
            outfile_prefix = self.add_model_id_to_prefix(self.outfile_prefix, each_reference.model_id)
            diagnostic_prefix = self.add_model_id_to_prefix(self.diagnostic_plot_prefix, each_reference.model_id)
            fsc_prefix = self.add_model_id_to_prefix('fsc.dat', each_reference.model_id)
        else:
            outfile_prefix = self.outfile_prefix
            diagnostic_prefix = self.diagnostic_plot_prefix
            fsc_prefix = 'fsc.dat'

        latest_reconstruction = self.generate_reconstruction_name_with_Angstrom_per_pixel_and_set_header(ref_cycle_id, 
        reconstructed_volume, outfile_prefix, pixelinfo.pixelsize, each_reference.helical_symmetry, 
        each_reference.point_symmetry)

        reconstructed_volume.write_image(latest_reconstruction)
        self.remove_intermediate_files_if_desired(each_reference.ref_file)
        reference_files[each_reference.model_id] = each_reference._replace(ref_file=latest_reconstruction)

        return latest_reconstruction, diagnostic_prefix, fsc_prefix, reference_files


    def determine_whether_is_last_cycle(self, start_ref_cycle_id, ref_cycle_id):
        current_cycle = ref_cycle_id - start_ref_cycle_id
        if current_cycle == self.iteration_count:
            is_last_cycle = True
        else:
            is_last_cycle = False

        return is_last_cycle


    def choose_reconstruction_stack_based_on_unbending(self, large_straightened_stack, large_reconstruction_stack):
        if large_straightened_stack is not None:
            reconstruction_stack = large_straightened_stack
        else:
            reconstruction_stack = large_reconstruction_stack
            
        return reconstruction_stack
    

    def cleanup_of_prj_stacks(self, prj_info):
        if prj_info.projection_stack is not None:
            if prj_info.projection_stack.startswith(self.tempdir):
                azimuthal_prj_stack = self.get_prj_stack_name_with_ending(prj_info.projection_stack, 'az')
                tilt_prj_stack = self.get_prj_stack_name_with_ending(prj_info.projection_stack, 'out')
    
                os.remove(azimuthal_prj_stack)
                os.remove(tilt_prj_stack)


# 
#     def perform_local_symmetry_refinement_based_on_power_spectra_matching(self, each_info, pixelinfo, ctf3d_avg_squared, each_reference, uncorrected_reconstruction, fsc_lines, mean_out_of_plane, exp_power, segment_size):
#         if self.refine_sym:
#             if self.rise_rot_or_pitch_unit_choice in ['rise/rotation']:
#                 pitch, no_unit = SegClassReconstruct().convert_rise_rotation_pair_to_pitch_unit_pair(each_reference.helical_symmetry)
#             else:
#                 pitch, no_unit = each_reference.helical_symmetry
#             grid_count = 25
#             rise_count = rot_count = int(np.sqrt(grid_count))
#             pitch_inc = 0.1
#             no_unit_inc = 0.01
#             sym_comb_grid = SegClassReconstruct().generate_rise_rotation_or_pitch_unitnumber_pairs_for_symmetry_grid(
#                 [pitch - (rise_count / 2) * pitch_inc, pitch + (rise_count / 2) * pitch_inc], pitch_inc, 
#                 [no_unit - (rot_count / 2) * no_unit_inc, no_unit + (rot_count / 2) * no_unit_inc], no_unit_inc)
#             print sym_comb_grid.shape
#             rises, rots = zip(*sym_comb_grid.ravel().tolist())
#             pixels = np.linspace(0, 1, len(fsc_lines.cylinder_masked))
#             fsc_px_0143 = self.get_resolution_closest_to_value(0.143, fsc_lines.cylinder_masked, pixels)
#             fourier_mask = model_circle(segment_size * fsc_px_0143 / 2.0, segment_size, segment_size)
#             if self.ctf_correction:
#                 corr_rec = self.perform_ctf_correction_on_volume(self.ctf_correction_type, 
#                     uncorrected_reconstruction, ctf3d_avg_squared, pixelinfo.pixelsize)
#             else:
#                 corr_rec = uncorrected_reconstruction
#             sym_combs = sym_comb_grid.ravel()
#             amp_ccs = np.zeros(len(sym_combs))
#             for each_id, (each_p, each_u) in enumerate(sym_combs):
#                 if self.rise_rot_or_pitch_unit_choice in ['rise/rotation']:
#                     each_sym_rise = SegClassReconstruct().convert_pitch_unit_pair_to_rise_rotation_pairs(each_p, each_u)
#                 else:
#                     each_sym_rise = each_p, each_u
#                 each_ref = each_reference._replace(helical_symmetry=each_sym_rise)
#                 reconstruction = self.generate_long_helix_volume(corr_rec, segment_size, 
#                     segment_size, each_ref.helical_symmetry, pixelinfo.pixelsize, each_ref.point_symmetry)
#                 sim_power, diagnostic_stack, projection_parameters, variance = self.generate_sim_power_from_reconstruction(each_info.resolution_aim, segment_size, 
#                     mean_out_of_plane, each_reference, pixelinfo, reconstruction)
#                 os.remove(diagnostic_stack)
#     #
#                 if each_info.resolution_aim in ['low', 'medium']:
#                     amp_cc_val = ccc(sim_power, exp_power, fourier_mask)
#                 elif each_info.resolution_aim in ['high', 'max']:
#                     amp_cc = SegClassReconstruct().compute_amplitude_correlation_between_sim_and_exp_power_spectrum(sim_power, exp_power)
#                     quarter, half, three_quarter, full = SegClassReconstruct().get_quarter_half_3quarter_nyquist_average_from_amp_correlation(amp_cc)
#                     if 0 <= fsc_px_0143 < 0.25:
#                         amp_cc_val = quarter
#                     elif 0.25 <= fsc_px_0143 < 0.5:
#                         amp_cc_val = half
#                     elif 0.5 <= fsc_px_0143 <= 0.75:
#                         amp_cc_val = three_quarter
#                     elif fsc_px_0143 >= 0.75:
#                         amp_cc_val = full
#                 amp_ccs[each_id] = amp_cc_val
#             
#             print amp_ccs
#     #                     max_sym = sym_combs[np.argmax(amp_ccs)]
#     #                     rise_count, rot_count = sym_comb_grid.shape
#             zoom_factor = 50.0
#             zoomed_amps = ndimage.zoom(amp_ccs.reshape((rise_count, rot_count)), zoom_factor)
#             zoomed_rises = np.linspace(rises[0], rises[-1], zoom_factor * rise_count)
#             zoomed_rots = np.linspace(rots[0], rots[-1], zoom_factor * rot_count)
#             zoomed_sym_combs = np.array([(each_rise, each_rot) for each_rise in zoomed_rises for each_rot in zoomed_rots])
#             max_sym = zoomed_sym_combs[np.argmax(zoomed_amps)]
#             if self.rise_rot_or_pitch_unit_choice in ['rise/rotation']:
#                 max_p, max_u = max_sym
#                 max_sym = SegClassReconstruct().convert_pitch_unit_pair_to_rise_rotation_pairs(max_p, max_u)
#             each_reference = each_reference._replace(helical_symmetry=max_sym)
#             print max_sym
#             if each_info.resolution_aim in ['low', 'medium']:
#                 reconstructed_volume = self.generate_long_helix_volume(corr_rec, reconstruction.get_xsize(), 
#                     reconstruction.get_zsize(), each_reference.helical_symmetry, pixelinfo.pixelsize, 
#                     each_reference.point_symmetry)
#             else:
#                 reconstructed_volume = corr_rec
#         else:
#             reconstructed_volume = self.perform_volume_operations_ctf_and_symmetrization(each_info.resolution_aim, 
#                 uncorrected_reconstruction, ctf3d_avg_squared, pixelinfo.pixelsize, each_reference)
#         return each_reference, reconstructed_volume, amp_cc, variance

    def reconstruct_volume(self, reference_files, lambda_sirt, start_ref_cycle_id, ref_cycle_id, each_info, pixelinfo,
    ctf3d_avg_squared, large_reconstruction_stack, selected_parameters, prj_info):
    
        if max([len(each_model) for each_model in selected_parameters]) == 0:
            self.if_no_selected_images_left_abort_refinement()

        rec_stack_info = SegClassReconstruct().make_rec_stack_info()
        for each_reference in reference_files:
            rec_stack = rec_stack_info(os.path.join(self.tempdir, 'rec_stack.hdf'), each_reference.ref_file,
            pixelinfo.alignment_size)
            
            if len(selected_parameters[each_reference.model_id]) > 1:
                if self.refine_symmetry:
                    each_zero_reference = each_reference._replace(helical_symmetry=(0.,0.))

                    uncorrected_reconstruction, alignment_parameters, symmetry_alignment_parameters, fsc_lines, lambda_sirt, Euler_angles_rec = \
                    self.apply_orientation_parameters_and_reconstruct_imposing_helical_symmetry(selected_parameters[each_reference.model_id],
                    ref_cycle_id, each_info.resolution_aim, large_reconstruction_stack, pixelinfo, each_zero_reference, 
                    self.stepsize, rec_stack, lambda_sirt, self.unbending)
                
                    each_reference, exp_power = \
                    self.perform_local_symmetry_refinement_based_on_power_spectra_matching(each_info, pixelinfo,
                    ctf3d_avg_squared, each_reference, uncorrected_reconstruction, fsc_lines, ref_cycle_id,
                    large_reconstruction_stack)
                else:
                    exp_power = None
                
                uncorrected_reconstruction, alignment_parameters, symmetry_alignment_parameters, fsc_lines, lambda_sirt, Euler_angles_rec = \
                self.apply_orientation_parameters_and_reconstruct_imposing_helical_symmetry(selected_parameters[each_reference.model_id],
                ref_cycle_id, each_info.resolution_aim, large_reconstruction_stack, pixelinfo, each_reference, 
                self.stepsize, rec_stack, lambda_sirt, self.unbending)
                
                reconstructed_volume = self.perform_volume_operations_ctf_and_symmetrization(each_info.resolution_aim,
                uncorrected_reconstruction, ctf3d_avg_squared, pixelinfo.pixelsize, each_reference)

                helical_error = self.perform_mean_ccc_evaluation_of_images_with_symmetry_related_projections(rec_stack,
                alignment_parameters, Euler_angles_rec, pixelinfo.helixwidthpix)
                
                self.log_helical_error(helical_error)
        
                latest_reconstruction, diagnostic_prefix, fsc_prefix, reference_files = \
                self.write_out_reconstruction_and_remove_reference(reference_files, ref_cycle_id, pixelinfo,
                each_reference, reconstructed_volume)
    
                self.log.plog(90 * (ref_cycle_id - start_ref_cycle_id + 0.9 - 1) / float(self.iteration_count) + 5)
                
                amp_cc, variance = self.summarize_each_bin_round_with_simulated_vs_experimental_images_and_powerspectra(
                each_info.resolution_aim, large_reconstruction_stack, latest_reconstruction, ref_cycle_id,
                each_reference, pixelinfo, diagnostic_prefix, prj_info, exp_power)
                
                is_last_cycle = self.determine_whether_is_last_cycle(start_ref_cycle_id, ref_cycle_id)

                self.write_out_fsc_line(fsc_lines, pixelinfo.pixelsize, fsc_prefix, ref_cycle_id)
                reference_files[each_reference.model_id] = reference_files[each_reference.model_id]._replace(fsc=fsc_lines)
            
                out_of_plane_dev = self.evaluate_alignment_parameters_and_summarize_in_plot(alignment_parameters,
                symmetry_alignment_parameters, fsc_lines, ref_cycle_id, each_reference, pixelinfo, diagnostic_prefix,
                each_info.resolution_aim)

                helical_error = helical_error._replace(out_of_plane_dev=out_of_plane_dev)

                self.enter_additional_ref_parameters_in_database(ref_cycle_id, symmetry_alignment_parameters,
                fsc_lines.cylinder_masked, amp_cc, variance, helical_error, pixelinfo.pixelsize,
                each_reference, is_last_cycle)
                
            else:
                reference_vol = EMData()
                reference_vol.read_image(each_reference.ref_file)
                latest_reconstruction, diagnostic_prefix, fsc_prefix, reference_files = \
                self.write_out_reconstruction_and_remove_reference(reference_files, ref_cycle_id, pixelinfo,
                each_reference, reference_vol)
            
        self.cleanup_of_prj_stacks(prj_info)

        return reference_files, lambda_sirt
    

    def print_fsc_to_diagnostic_plot(self, reference_files, pixelinfo, ref_cycle_id, each_ref_id, each_reference,
    fsc_lines, fsc_prefix):
        fsc_plot = DiagnosticPlot()
        column_count = 1
        row_count = 2
        ax1 = fsc_plot.plt.subplot2grid((row_count, column_count), (0, 0), rowspan=1, colspan=1)
        ax2 = fsc_plot.plt.subplot2grid((row_count, column_count), (1, 0), rowspan=1, colspan=1)

        ax1.set_title('FSC from independent half-set halves (top) and independent cross-FSC (bottom)', fontsize=10)
        fsc_plot.set_fontsize_to_all_ticklabels_of_subplots([ax1, ax2], font_size=6)
        ax1, res_cutoff = self.plot_fsc_lines(ax1, reference_files[2 * each_ref_id].fsc, pixelinfo.pixelsize)
        ax1, res_cutoff = self.plot_fsc_lines(ax1, reference_files[2 * each_ref_id + 1].fsc, pixelinfo.pixelsize)

        ax2, (res_0143, res_05) = self.plot_fsc_lines(ax2, fsc_lines, pixelinfo.pixelsize)
        fsc_name = self.write_out_fsc_line(fsc_lines, pixelinfo.pixelsize, fsc_prefix, ref_cycle_id)
        plotname = os.path.splitext(fsc_name)[0] + os.path.splitext(self.diagnostic_plot_prefix)[-1]
        fsc_plot.fig.savefig(plotname, dpi=600)

        table_data = [['Resolution at FSC (0.5/0.143) (Angstrom)', res_05, res_0143]]
        self.log.ilog('The cross FSC from independent half-set refinements was determined:\n{0}'.format(tabulate(table_data)))

        resolution = SegmentExam().make_oneoverres(fsc_lines[0], pixelinfo.pixelsize)
        msg = tabulate(zip(resolution.tolist(), fsc_lines.unmasked, fsc_lines.cylinder_masked), 
            ['resolution (1/Angstrom)', 'FSC (unmasked)', 'FSC (cylinder-masked)'])
        self.log.ilog('\n{0}'.format(msg))
            
        return res_0143, res_05


    def update_highest_fsc_database(self, ref_cycle_id, model_fscs, model_resolution):
        if model_resolution != []:
            temp_ref_db = self.copy_ref_db_to_tempdir(ref_cycle_id)
            ref_session = SpringDataBase().setup_sqlite_db(refine_base, temp_ref_db)
    
            model_0143, model_05 = zip(*model_resolution)
            fsc_to_be_entered = model_fscs[np.argmax(model_0143)]
    
            last_cycle = ref_session.query(RefinementCycleTable).order_by(desc(RefinementCycleTable.id)).first()
            last_cycle.fsc_0143 = model_0143[np.argmax(model_0143)]
            last_cycle.fsc_05 = model_05[np.argmax(model_0143)]
            last_cycle.fsc = fsc_to_be_entered
            last_cycle.fsc_split = True
    
            ref_session.merge(last_cycle)
            ref_session.commit()
    
            shutil.copy(temp_ref_db, 'refinement{0:03}.db'.format(ref_cycle_id))
            os.remove(temp_ref_db)


    def compute_cross_fsc_and_write_out_merged_volume(self, reference_files, resolution_aim, pixelinfo, ref_cycle_id):
        model_fscs = []
        model_resolution = []
        for each_ref_id in list(range(len(reference_files) / 2)):
            each_reference = reference_files[2 * each_ref_id]
            rec_odd = reference_files[2 * each_ref_id].ref_file
            rec_even = reference_files[2 * each_ref_id + 1].ref_file
            
            first_rec = EMData()
            second_rec = EMData()
            first_rec.read_image(rec_odd)
            second_rec.read_image(rec_even)
            
            helical_symmetry = each_reference.helical_symmetry
            rotational_sym = each_reference.rotational_symmetry
            point_symmetry = each_reference.point_symmetry

            reconstructed_volume, fsc_lines = self.compute_fsc_on_volumes_from_half_the_dataset(resolution_aim, first_rec,
            second_rec, pixelinfo, helical_symmetry, rotational_sym)
            
            if len(reference_files) > 2:
                outfile_prefix = self.add_model_id_to_prefix(self.outfile_prefix, each_reference.model_id)
                fsc_prefix = self.add_model_id_to_prefix('fsc_indpndnt.dat', each_reference.model_id)
            else:
                outfile_prefix = self.outfile_prefix
                fsc_prefix = 'fsc_indpndnt.dat'

            latest_reconstruction = \
            self.generate_reconstruction_name_with_Angstrom_per_pixel_and_set_header(ref_cycle_id, reconstructed_volume,
            outfile_prefix, pixelinfo.pixelsize, helical_symmetry, point_symmetry)
            reconstructed_volume.write_image(latest_reconstruction)

            res_cutoffs = self.print_fsc_to_diagnostic_plot(reference_files, pixelinfo, ref_cycle_id, each_ref_id, each_reference,
            fsc_lines, fsc_prefix)
            
            model_fscs.append(fsc_lines.cylinder_masked)
            model_resolution.append(res_cutoffs)

            reference_files[2 * each_ref_id] = each_reference._replace(fsc=fsc_lines)
            reference_files[2 * each_ref_id + 1] = reference_files[2 * each_ref_id + 1]._replace(fsc=fsc_lines)

        self.update_highest_fsc_database(ref_cycle_id, model_fscs, model_resolution)

        return reference_files


    def bin_including_copies_of_frames_if_requested(self, image_list, ori_large_segment_stack, ori_alignment_size,
    ori_helixwidthpix, ori_pixelsize, each_info):
        if self.frame_motion_corr and self.frame_avg_window > 2:
            frames_list = zip(*zip(*image_list)[1])
            large_binned_segment_stack = os.path.join(self.tempdir, ori_large_segment_stack)
            for each_frame_id, each_frame_image_ids in enumerate(frames_list):
                frame_stack_path = self.get_frame_stack_path(ori_large_segment_stack, each_frame_id)

                large_binned_frame_stack, segment_size_pix, helixwidthpix, pixelsize = \
                SegmentExam().apply_binfactor(each_info.bin_factor, ori_large_segment_stack, ori_alignment_size,
                ori_helixwidthpix, ori_pixelsize, each_frame_image_ids, frame_stack_path)
            
            os.rename(self.get_frame_stack_path(ori_large_segment_stack, 0), large_binned_segment_stack)
            os.symlink(large_binned_segment_stack, self.get_frame_stack_path(ori_large_segment_stack, 0))
        else:
            large_binned_segment_stack, segment_size_pix, helixwidthpix, pixelsize = \
            SegmentExam().apply_binfactor(each_info.bin_factor, ori_large_segment_stack, ori_alignment_size,
            ori_helixwidthpix, ori_pixelsize, image_list, os.path.join(self.tempdir, ori_large_segment_stack))

        return large_binned_segment_stack, segment_size_pix, helixwidthpix


    def perform_iterative_projection_matching_and_3d_reconstruction(self):
        OpenMpi().setup_and_start_mpi_version_if_demanded(self.mpi_option, self.feature_set, self.cpu_count)
        
        required_byte, info_series = self.estimate_required_tmp_disk_space()
        self.tempdir = Temporary().mktmpdir(self.temppath, required_byte)
        
        large_segment_stack, masked_segment_stack, image_list, reference_files, lambda_sirt, start_ref_cycle_id = \
        self.pre_cycle_setup(info_series)
        
        ori_large_segment_stack, ori_reference_files, ori_alignment_size, ori_helixwidthpix, ori_pixelsize = \
        self.define_original_input_values_before_binning(large_segment_stack, reference_files)
        
        self.log.plog(5)
        
        ref_cycle_id, unbending_info, large_straightened_stack = \
        self.setup_dummies_for_iteration(start_ref_cycle_id)
        
        for each_index, each_info in enumerate(info_series):
            large_binned_segment_stack, segment_size_pix, helixwidthpix = \
            self.bin_including_copies_of_frames_if_requested(image_list, ori_large_segment_stack, ori_alignment_size,
            ori_helixwidthpix, ori_pixelsize, each_info)
            
            reference_files, ctf3d_avg_squared, pixelinfo = self.precycle_setup_before_binning(info_series,
            large_binned_segment_stack, reference_files, ori_reference_files, ori_pixelsize, each_index, ref_cycle_id,
            segment_size_pix, helixwidthpix)
        
            for each_iteration_number in list(range(1, each_info.iteration_count + 1)):
                ref_cycle_id += 1
                
                mask_params, previous_params, prj_info, reference_files = \
                self.project_including_masking_and_filtering(required_byte, info_series, reference_files,
                start_ref_cycle_id, ref_cycle_id, each_info, each_iteration_number, pixelinfo)
        
                masked_segment_stack, previous_params, unbending_info, large_straightened_stack, \
                large_reconstruction_stack = self.mask_and_window_and_unbend_if_required(info_series,
                masked_segment_stack, start_ref_cycle_id, ref_cycle_id, each_info, large_binned_segment_stack,
                each_iteration_number, mask_params, previous_params, unbending_info, large_straightened_stack,
                pixelinfo)
                    
                if self.resume_refinement_option and not self.reference_option and each_iteration_number == 1:
                    current_parameters = previous_params
                    translation_step = 1.0
                else:
                    current_parameters, translation_step, prj_info = \
                    self.perform_coarse_and_fine_projection_matching(each_info, masked_segment_stack, prj_info,
                    previous_params, pixelinfo.alignment_size)
            
                large_binned_segment_stack = large_reconstruction_stack

                self.log.plog(90 * (ref_cycle_id - start_ref_cycle_id + 0.7 - 1) / float(self.iteration_count) + 5)
                
                selected_parameters = self.select_segments_based_on_specified_criteria(current_parameters,
                unbending_info, translation_step, ref_cycle_id, each_info, pixelinfo, reference_files)
                
                reconstruction_stack = self.choose_reconstruction_stack_based_on_unbending(large_straightened_stack,
                large_reconstruction_stack)
                    
                reference_files, lambda_sirt = self.reconstruct_volume(reference_files, lambda_sirt,
                start_ref_cycle_id, ref_cycle_id, each_info, pixelinfo, ctf3d_avg_squared,
                reconstruction_stack, selected_parameters, prj_info)
                
                self.log.plog(90 * (ref_cycle_id - start_ref_cycle_id) / float(self.iteration_count) + 5)
                
                aim_ref = self.get_aim_dict()
                if self.halfset_refinement and aim_ref[each_info.resolution_aim] >= aim_ref[self.halfset_start]:
                    reference_files = self.compute_cross_fsc_and_write_out_merged_volume(reference_files, each_info,
                    pixelinfo, ref_cycle_id)

                if self.layer_line_filter and not self.unbending:
                    os.remove(large_straightened_stack)
        
            if self.unbending:
                os.remove(large_straightened_stack)
                
            os.remove(large_reconstruction_stack)
        
        self.update_persistence_length_in_spring_db()

        os.remove(large_segment_stack)
        self.clean_up_temporary_large_stack(large_segment_stack, masked_segment_stack)
        if self.ctf_correction:
            os.remove(ctf3d_avg_squared)
        os.rmdir(self.tempdir)
        
        self.log.endlog(self.feature_set)
            
def main():
    # Option handling
    parset = SegmentRefine3dPar()
    mergeparset = OptHandler(parset)

    ######## Program
    stack = SegmentRefine3d(mergeparset)
    stack.perform_iterative_projection_matching_and_3d_reconstruction()


if __name__ == '__main__':
    main()



# Author: Carsten Sachse 21-Sep-2010
# Copyright: EMBL (2010 - 2018)
# License: see license.txt for details
""" 
Program to evaluate scanner performance of scanner by measuring deviation from 45 degree 
line to determine CCD curvature and pincushion parameter.
"""

from EMAN2 import Util
from spring.csinfrastr.csfeatures import Features
from spring.csinfrastr.cslogger import Logger
from spring.csinfrastr.csproductivity import DiagnosticPlot
from spring.csinfrastr.csreadinput import OptHandler
from spring.micprgs.scandotfit import ScanDotFit
from spring.micprgs.scansplit import Micrograph
import numpy as np

class ScanLineFitPar:
    """
    Class to initiate default dictionary with input parameters including help and range values and 
    status dictionary
    """
    def __init__(self):
        # package/program identity
        self.package = 'emspring'
        self.progname = 'scanlinefit'
        self.proginfo = __doc__
        self.code_files = [self.progname]

        self.scanlinefit_features = Features()
        self.feature_set = self.scanlinefit_features.setup(self)
        
        self.define_parameters_and_their_properties()
        self.define_program_states()
    
    def define_parameters_and_their_properties(self):
        self.feature_set = self.scanlinefit_features.set_inp_micrograph(self.feature_set)
        self.feature_set = self.scanlinefit_features.set_output_plot(self.feature_set, self.progname + '_diag.pdf')
    
        self.feature_set = self.set_integration_width(self.feature_set)
        self.feature_set = self.set_topleft_coordinates(self.feature_set)
        self.feature_set = self.set_bottomright_coordinates(self.feature_set)
        
    def define_program_states(self):
        self.feature_set.program_states['readmic']='Read input micrograph'
        self.feature_set.program_states['getline']='Get line coordinates from micrograph'
        self.feature_set.program_states['fitcube']='Fit extracted coordinates to cubic function'
        self.feature_set.program_states['visfit']='Visualize extracted line and fitted function'


    def set_integration_width(self, feature_set):
        inp3 = 'Width of integration in pixels'
        feature_set.parameters[inp3] = int(9)
        feature_set.hints[inp3] = 'Use default value: number of pixels used for line determination - otherwise use ' +\
        'with caution.'
        feature_set.properties[inp3] = feature_set.Range(1, 500, 1)
        feature_set.level[inp3]='expert'
        
        return feature_set


    def set_topleft_coordinates(self, feature_set):
        inp4 = 'Topleft coordinates of line'
        feature_set.parameters[inp4] = tuple((180, 2302))
        feature_set.hints[inp4] = 'Integer pair of coordinates (x,y).'
        feature_set.properties[inp4] = feature_set.Range(0, 30000, 1)
        feature_set.level[inp4] = 'intermediate'
        
        return feature_set


    def set_bottomright_coordinates(self, feature_set):
        inp5 = 'Bottomright coordinates of line'
        feature_set.parameters[inp5] = tuple((1889, 215))
        feature_set.hints[inp5] = 'Integer pair of coordinates (x,y).'
        feature_set.properties[inp5] = feature_set.Range(0, 30000, 1)
        feature_set.level[inp5]='intermediate'
        
        return feature_set
    

class ScanLineFit(object):
    """
    * Class that holds functions for determining deviation of an ideal 45 degree line

    * __init__ Function to read in the entered parameter dictionary and load micrograph

    #. Usage: ScanLineFit(pardict)
    #. Input: pardict = OrderedDict of program parameters
    """
    def __init__(self, parset = None):
        self.log = Logger()
        if parset is not None:
            self.feature_set = parset
            p = self.feature_set.parameters

            self.infile = p['Micrograph']
            self.outfile = p['Diagnostic plot']
            self.topleft = p['Topleft coordinates of line']
            self.bottomright = p['Bottomright coordinates of line']
            self.iwidth = p['Width of integration in pixels']

            self.img, self.nx, self.ny = Micrograph().readmic(self.infile)
        
    def getline(self, iwidth = None, topleft = None, bottomright = None):
        """
        * Function to determine 45 degree line and their corresponding x, y coordinates using\
                        rough starting coordindates and refine by center-of-gravity measurements
        
        #. Usage: output1 = getline(iwidth, topleft, bottomright)
        #. Input: iwidth = integration width, topleft = topleft coordinates of line, bottomright = \
                        bottom right coordinates of line
        #. Output: 1. array of x and y
        """
        self.log.fcttolog()
        if iwidth is None: iwidth = self.iwidth
        if topleft is None: topleft = self.topleft
        if bottomright is None: bottomright = self.bottomright

        df = ScanDotFit()
        slope, icept = df.pttoline(topleft, bottomright)

        if topleft[1] < bottomright[1]:
            startrow, endrow = topleft[1] - iwidth, bottomright[1] + iwidth
            #startrow, endrow = topleft[1] - integration_width, bottomright[1] - integration_width
        else:
            startrow, endrow = bottomright[1] + iwidth, topleft[1] - iwidth
        self.log.dlog('startrow: %g, endrow: %g' %(startrow, endrow))

        self.linex = [int(round((eachRow-icept)/slope)) for eachRow in range(startrow, endrow)]
        self.liney = [eachRow for eachRow in range(startrow, endrow)]

        def cog_row(img, nx, ny, iwidth, linex, liney, k):
            xoffset = int(-nx/2 + linex[k] + 0.5)
            yoffset = int(-ny/2 + liney[k] + 0.5)
            rowimg = Util.window(img, iwidth, 1, 1, xoffset, yoffset, 0)
            #rowimg.write_image('rowstack.spi', k)
            xshift = rowimg.phase_cog()
            self.log.ilog('Offset from ideal line: %g' %xshift[0])
            linexx = xshift[0] + linex[k]
            self.log.ilog('Pixel density added to array: idealcol %g, measuredcol %g, row %g' %
                          (linex[k], linexx, liney[k]))
            return linexx

        self.linex = [cog_row(self.img, self.nx, self.ny, iwidth, self.linex, self.liney, k) \
                        for k in range(len(self.linex))]
        return self.linex, self.liney

    def fitcube(self, linex = None, liney = None):
        """
        * Function to perform cubic polynomial fit to binary line and determine 

        #. Usage: polyvar, fitliney3, fitliney1 = fitcube(linex, liney)
        #. Input: linex = x and liney = y array of thresholded 45 degree line
        #. Output: list of a, b, c, d (a, b, c = measure of CCD curvature, d = pincushion \
                        or barrel distortion), fitliney3 = computed array of cubic fit, \
                        fitliney1 = computed array of linear fit
        """
        self.log.fcttolog()
        if linex is None: linex = self.linex
        if liney is None: liney = self.liney

        self.polyvar = np.polyfit(linex, liney, 3)
        polyline = np.polyfit(linex, liney, 1)
        self.log.ilog('The threshold pixel line has been fitted with a cubic function:')
        self.log.ilog('Fit: %g*x^3 + %g*x^2 + %g*x + %g' %
                      (self.polyvar[0], self.polyvar[1], self.polyvar[2], self.polyvar[3]))
        # compute line according to fit variables
        self.fitliney3 = np.polyval(self.polyvar, linex)
        self.fitliney1 = np.polyval(polyline, linex)

        return self.polyvar, self.fitliney3, self.fitliney1

    def visfit(self, linex = None, liney = None, polyvar = None, fitliney3 = None, fitliney1 = None):
        """
        * Function to visualize 45 degree line including cubic polynomial fit using matplotlib

        #. Usage: output = visfit(linex, liney, polyvar, fitliney3, fitliney1)
        #. Input: linex = x and liney = y array of thresholded 45 degree line, polyvar = list \
                        of a, b, c, d (a, b, c = measure of CCD curvature, d = pincushion or \
                        barrel distortion fitliney3 = computed array of cubic fit, fitliney1 \
                        = computed array of linear fit
        #. Output: output plot to saved to PDF, SVG, or PNG, TIF, JPG format
        """
        import matplotlib.font_manager as font_manager
        from math import sqrt
        self.log.fcttolog()

        if linex is None: linex = self.linex
        if liney is None: liney = self.liney
        if polyvar is None: polyvar = self.polyvar
        if fitliney3 is None: fitliney3 = self.fitliney3
        if fitliney1 is None: fitliney1 = self.fitliney1

        distance = sqrt((self.linex[-1]-self.linex[0])**2+(self.liney[-1]-self.liney[0])**2)
        self.log.ilog('Line distance from center in pixel: %g' %(distance/2))
        magpin = polyvar[0]*(distance/2)**3
        magccd = polyvar[1]*(distance/2)**2
        self.log.ilog('Magnitude pincushion: %g, CCD curvature: %g (pixel)' %(magpin, magccd))

        scanlinefit_plot = DiagnosticPlot()
        self.fig = scanlinefit_plot.add_header_and_footer(self.feature_set)

        ax1 = self.fig.add_subplot(111)
        ax1.plot(linex, liney, 'x', markeredgewidth=0.02, markersize=0.4, label = 'Thresholded line')
        ax1.plot(linex, fitliney3, linewidth = .02, label = 'Fit: %g*x^3 + %g*x^2 + %g*x + %g' 
                 %(polyvar[0], polyvar[1], polyvar[2], polyvar[3]))
        ax1.plot(linex, fitliney1, linewidth = .02, label = 'Linear fit')
        ax1.set_title('45 Degree line with fitted cubic function')
        ax1.text(1.3, 0.25, 'magnitude pincushion: %g, CCD curvature: %g (pixel)' %(magpin, magccd), fontsize=8)
        ax1.legend(loc=(.2,0.85) , prop=font_manager.FontProperties(size='x-small'))
        ax1 = ScanDotFit().vismic(self.nx, self.ny, ax1)
        self.fig.savefig(self.outfile)
        return self.fig

    def perform_scanlinefit(self):
        self.log.plog(10)
        self.getline()
        self.log.plog(40)
        self.fitcube()
        self.log.plog(70)
        self.visfit()
        self.log.endlog(self.feature_set)
        
def main():
    # Option handling
    parset = ScanLineFitPar()
    mergeparset = OptHandler(parset)
    ######## Program
    mic = ScanLineFit(mergeparset)
    mic.perform_scanlinefit()

if __name__ == '__main__':
    main()
